# Copyright(C) 2010-2012  Christophe Benz, Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from collections import OrderedDict
from copy import copy
import getpass
import logging
import shlex
import subprocess
from subprocess import check_output
import sys
import os
from tempfile import NamedTemporaryFile
import warnings

from rich.progress import Progress, TaskProgressColumn, BarColumn, TextColumn

from woob.capabilities import UserError
from woob.capabilities.account import CapAccount, Account, AccountRegisterError
from woob.core.backendscfg import BackendAlreadyExists
from woob.core.repositories import IProgress
from woob.exceptions import (
    BrowserUnavailable, BrowserIncorrectPassword, BrowserForbidden,
    BrowserSSLError, BrowserQuestion, BrowserHTTPSDowngrade,
    ModuleInstallError, ModuleLoadError, ActionNeeded, CaptchaQuestion,
    NeedInteractiveFor2FA,
)
from woob.tools.application.pretty import colored, NC, BOLD
from woob.tools.value import Value, ValueBool, ValueFloat, ValueInt, ValueBackendPassword
from woob.tools.misc import to_unicode, classproperty

from .base import Application, MoreResultsAvailable


__all__ = ['ConsoleApplication', 'BackendNotGiven']


class BackendNotGiven(Exception):
    def __init__(self, id, backends):
        self.id = id
        self.backends = sorted(backends)
        super().__init__(
            'Please specify a backend to use for this argument (%s@backend_name). '
            'Availables: %s.' % (id, ', '.join(name for name, backend in backends))
        )


class BackendNotFound(Exception):
    pass


class ConsoleProgress(IProgress):
    def __init__(self, app):
        self.app = app
        self._progress = None
        self._progress_task = None

    def progress(self, percent, message):
        if not self._progress:
            self._progress = Progress(
                TaskProgressColumn(),
                BarColumn(),
                TextColumn("[progress.description]{task.description}")
            )
            self._progress_task = self._progress.add_task(message, total=1)
            self._progress.start()

        self._progress.update(self._progress_task, advance=percent, description=message)

        if percent == 1:
            self._progress.stop()
            self._progress = None

    def error(self, message):
        if self._progress:
            self._progress.stop()

        print(colored(f'ERROR: {message}', 'red'), file=self.app.stderr)

        if self._progress:
            self._progress.start()

    def prompt(self, message: str, default: bool = True) -> bool:
        try:
            if self._progress:
                self._progress.stop()
            return self.app.ask(message, default=default)
        finally:
            if self._progress:
                self._progress.start()


class ConsoleApplication(Application):
    """
    Base application class for CLI applications.
    """

    CAPS = None

    @classproperty
    def BOLD(self):
        """
        .. deprecated:: 3.6
            This attribute is deprecated, use :attr:`woob.tools.application.pretty.BOLD` instead.
            That's also better to use :func:`woob.tools.application.pretty.colored`.
        """
        warnings.warn(
            'Use woob.tools.application.pretty.BOLD instead.\n'
            'That\'s also better to use woob.tools.application.pretty.colored.',
            DeprecationWarning,
            stacklevel=2
        )
        return BOLD

    @classproperty
    def NC(self):
        """
        .. deprecated:: 3.6
            This attribute is deprecated, use :attr:`woob.tools.application.pretty.NC` instead.
            That's also better to use :func:`woob.tools.application.pretty.colored`.
        """
        warnings.warn(
            'Use woob.tools.application.pretty.NC instead.\n'
            'That\'s also better to use woob.tools.application.pretty.colored.',
            DeprecationWarning,
            stacklevel=2
        )
        return NC

    def __init__(self, option_parser=None):
        super().__init__(option_parser)
        self.woob.requests.register('login', self.login_cb)
        self.enabled_backends = set()
        self._parser.add_option(
            '--auto-update',
            action='store_true',
            help='Automatically check for updates when a bug in a module is encountered'
        )

    def login_cb(self, backend_name, value):
        return self.ask('[%s] %s' % (backend_name,
                        value.label),
                        masked=True,
                        default='',
                        regexp=value.regexp)

    def unload_backends(self, *args, **kwargs):
        unloaded = self.woob.unload_backends(*args, **kwargs)
        for backend in unloaded.values():
            try:
                self.enabled_backends.remove(backend)
            except KeyError:
                pass
        return unloaded

    def is_module_loadable(self, info):
        return self.CAPS is None or info.has_caps(self.CAPS)

    def load_backends(self, *args, **kwargs):
        if 'errors' in kwargs:
            errors = kwargs['errors']
        else:
            kwargs['errors'] = errors = []
        ret = super().load_backends(*args, **kwargs)

        for err in errors:
            print('Error(%s): %s' % (err.backend_name, err), file=self.stderr)
            if self.ask('Do you want to reconfigure this backend?', default=True):
                self.edit_backend(err.backend_name)
                self.load_backends(names=[err.backend_name])

        for name, backend in ret.items():
            self.enabled_backends.add(backend)

        self.check_loaded_backends()

        return ret

    def check_loaded_backends(self, default_config=None):
        if len(self.enabled_backends) == 0:
            print('There is currently no configured backend for %s' % self.APPNAME)
            self.prompt_create_backends(default_config)

        return True

    def prompt_create_backends(self, default_config=None):
        r = ''
        while True:
            modules = []
            print('\nAvailable modules:')
            for name, info in sorted(self.woob.repositories.get_all_modules_info().items()):
                if not self.is_module_loadable(info):
                    continue
                modules.append(name)
                loaded = ' '
                for bi in self.woob.iter_backends():
                    if bi.NAME == name:
                        if loaded == ' ':
                            loaded = 'X'
                        elif loaded == 'X':
                            loaded = 2
                        else:
                            loaded += 1
                print(
                    '%s) [%s] %s%s' % (
                        colored('%2d' % len(modules), 'white', attrs=['bold']),
                        colored(str(loaded), attrs=['bold']),
                        colored('%-20s' % name, 'magenta', attrs=['bold']),
                        colored(info.description, 'green')
                    )
                )

            print(
                ' %s) %s                 %s' % (
                    colored('a', 'white', attrs=['bold']),
                    colored('--all--', attrs=['bold']),
                    colored('install all backends', 'green')
                )
            )
            print(
                ' %s) %s\n' % (
                    colored('q', 'white', attrs=['bold']),
                    colored('--stop--', attrs=['bold'])
                )
            )
            r = self.ask('Select a backend to create', default='')

            if r in ('q', ''):
                break
            if r == 'a':
                try:
                    for name in modules:
                        if name in [b.NAME for b in self.woob.iter_backends()]:
                            continue
                        inst = self.add_backend(name, name, default_config)
                        if inst:
                            self.load_backends(names=[inst])
                except (KeyboardInterrupt, EOFError):
                    print('\nAborted.')
                else:
                    break
                continue

            if r.isdigit():
                i = int(r) - 1
                if i < 0 or i >= len(modules):
                    print(f'Error: {r} is not a valid choice', file=self.stderr)
                    continue
                name = modules[i]
            elif r in modules:
                name = r
            else:
                print(f'Error: {r} is not a valid choice', file=self.stderr)
                continue

            try:
                inst = self.add_backend(name, name, default_config)
                if inst:
                    self.load_backends(names=[inst])
            except (KeyboardInterrupt, EOFError):
                print('\nAborted.')

    def _handle_options(self):
        self.load_default_backends()

    def load_default_backends(self):
        """
        By default loads all backends.

        Applications can overload this method to restrict backends loaded.
        """
        if len(self.STORAGE) > 0:
            self.load_backends(self.CAPS, storage=self.create_storage())
        else:
            self.load_backends(self.CAPS)

    @classmethod
    def run(cls, args=None):
        try:
            super().run(args)
        except BackendNotFound as e:
            print(f'Error: Backend "{e}" not found.')
            sys.exit(1)

    def do(self, function, *args, **kwargs):
        if 'backends' not in kwargs:
            kwargs['backends'] = self.enabled_backends
        return self.woob.do(function, *args, **kwargs)

    def parse_id(self, _id, unique_backend=False):
        try:
            _id, backend_name = _id.rsplit('@', 1)
        except ValueError:
            backend_name = None
        backends = [(b.name, b) for b in self.enabled_backends]
        if unique_backend and not backend_name:
            if len(backends) == 1:
                backend_name = backends[0][0]
            else:
                raise BackendNotGiven(_id, backends)
        if backend_name is not None and backend_name not in dict(backends):
            # Is the backend a short version of a real one?
            found = False
            for key in dict(backends):
                if backend_name in key:
                    # two choices, ambiguous command
                    if found:
                        raise BackendNotFound(backend_name)
                    else:
                        found = True
                        _back = key
            if found:
                return _id, _back
            raise BackendNotFound(backend_name)
        return _id, backend_name

    # user interaction related methods

    def register_backend(self, name, ask_add=True):
        try:
            backend = self.woob.modules_loader.get_or_load_module(name)
        except ModuleLoadError:
            backend = None

        if not backend:
            print('Backend "%s" does not exist.' % name, file=self.stderr)
            return 1

        if not backend.has_caps(CapAccount) or backend.klass.ACCOUNT_REGISTER_PROPERTIES is None:
            print('You can\'t register a new account with %s' % name, file=self.stderr)
            return 1

        account = Account()
        account.properties = {}
        if backend.website:
            website = 'on website %s' % backend.website
        else:
            website = 'with backend %s' % backend.name
        while True:
            asked_config = False
            for key, prop in backend.klass.ACCOUNT_REGISTER_PROPERTIES.items():
                if not asked_config:
                    asked_config = True
                    print('Configuration of new account %s' % website)
                    print('-----------------------------%s' % ('-' * len(website)))
                p = copy(prop)
                p.set(self.ask(prop, default=account.properties[key].get() if (key in account.properties) else prop.default))
                account.properties[key] = p
            if asked_config:
                print('-----------------------------%s' % ('-' * len(website)))
            try:
                backend.klass.register_account(account)
            except AccountRegisterError as e:
                print('%s' % e)
                if self.ask('Do you want to try again?', default=True):
                    continue
                else:
                    return None
            else:
                break
        backend_config = {}
        for key, value in account.properties.items():
            if key in backend.config:
                backend_config[key] = value.get()

        if ask_add and self.ask('Do you want to add the new register account?', default=True):
            return self.add_backend(name, name, backend_config, ask_register=False)

        return backend_config

    def install_module(self, minfo):
        try:
            self.woob.repositories.install(minfo, ConsoleProgress(self))
        except ModuleInstallError as e:
            print('Unable to install module "%s": %s' % (minfo.name, e), file=self.stderr)
            return False

        return True

    def edit_backend(self, name, params=None):
        return self.add_backend(name, name, params, True)

    def add_backend(self, module_name, backend_name, params=None, edit=False, ask_register=True):
        if params is None:
            params = {}

        module = None
        config = None
        try:
            if not edit:
                minfo = self.woob.repositories.get_module_info(module_name)
                if minfo is None:
                    raise ModuleLoadError(module_name, 'Module does not exist')

                if not minfo.is_installed() and not self.install_module(minfo):
                    return

                module = self.woob.modules_loader.get_or_load_module(module_name)
                config = module.config
            else:
                module_name, items = self.woob.backends_config.get_backend(backend_name)
                module = self.woob.modules_loader.get_or_load_module(module_name)
                items.update(params)
                params = items
                config = module.config.load(self.woob, module_name, backend_name, params, nofail=True)
        except ModuleLoadError as e:
            print('Unable to load module "%s": %s' % (module_name, e), file=self.stderr)
            return

        # ask for params non-specified on command-line arguments
        asked_config = False
        for key, value in config.items():
            if value.transient:
                continue

            if not asked_config:
                asked_config = True
                print('')
                print('Configuration of backend %s' % module.name)
                print('-------------------------%s' % ('-' * len(module.name)))
            if key not in params or edit:
                params[key] = self.ask(value, default=params[key] if (key in params) else value.default)
            else:
                print('[%s] %s: %s' % (key, value.description, '(masked)' if value.masked else to_unicode(params[key])))
        if asked_config:
            print('-------------------------%s' % ('-' * len(module.name)))

        i = 2
        while not edit and self.woob.backends_config.backend_exists(backend_name):
            if not self.ask('Backend "%s" already exists. Add a new one for module %s?' % (backend_name, module.name), default=False):
                return 1

            backend_name = backend_name.rstrip('0123456789')
            while self.woob.backends_config.backend_exists('%s%s' % (backend_name, i)):
                i += 1
            backend_name = self.ask('Please give new backend name', default='%s%s' % (backend_name, i), regexp=r'^[\w\-_]+$')

        if edit:
            self.woob.backends_config.edit_backend(backend_name, params)
        else:
            try:
                self.woob.backends_config.add_backend(backend_name, module.name, params)
            except BackendAlreadyExists:
                print('Backend "%s" already exists.' % backend_name, file=self.stderr)
                return None

        print('Backend "%s" successfully %s.' % (backend_name, 'edited' if edit else 'added'))
        return backend_name

    def ask(self, question, default=None, masked=None, regexp=None, choices=None, tiny=None):
        """
        Ask a question to user.

        :param question: text displayed
        :type question: str
        :param default: optional default value
        :type default: str
        :param masked: if True, do not show typed text
        :type masked: bool
        :param regexp: text must match this regexp
        :type regexp: str
        :param choices: choices to do
        :type choices: list
        :param tiny: ask for the (small) value of the choice
        :type tiny: bool
        :returns: text by user
        :rtype: str
        """

        if isinstance(question, Value):
            v = copy(question)
            if default is not None:
                v.default = default
            if masked is not None:
                v.masked = masked
            if regexp is not None:
                v.regexp = regexp
            if choices is not None:
                v.choices = choices
            if tiny is not None:
                v.tiny = tiny
        else:
            if isinstance(default, bool):
                klass = ValueBool
            elif isinstance(default, float):
                klass = ValueFloat
            elif isinstance(default, int):
                klass = ValueInt
            else:
                klass = Value

            v = klass(label=question, default=default, masked=masked, regexp=regexp, choices=choices, tiny=tiny)

        question = v.label
        if v.description and v.description != v.label:
            question = '%s: %s' % (question, v.description)
        if v.id:
            question = '[%s] %s' % (v.id, question)

        if isinstance(v, ValueBackendPassword):
            print(question + ':')
            question = v.label
            choices = OrderedDict()
            choices['c'] = 'Run an external tool during backend load'
            if not v.noprompt:
                choices['p'] = 'Prompt value when needed (do not store it)'
            choices['s'] = 'Store value in config'

            if v.default == '' and not v.noprompt:
                default = 'p'
            else:
                default = 's'

            r = self.ask('*** How do you want to store it?', choices=choices, tiny=True, default=default)
            if r == 'p':
                return ''
            if r == 'c':
                print('Enter the shell command that will print the required value on the standard output')
                while True:
                    cmd = self.ask('')
                    try:
                        check_output(cmd, shell=True)  # nosec: this is intended
                    except subprocess.CalledProcessError as e:
                        print('%s' % e)
                    else:
                        return '`%s`' % cmd

        aliases = {}
        if isinstance(v, ValueBool):
            question = '%s (%s/%s)' % (question, 'Y' if v.default else 'y', 'n' if v.default else 'N')
        elif v.choices:
            if v.tiny is None:
                v.tiny = True
                for key in v.choices:
                    if len(key) > 5 or ' ' in key:
                        v.tiny = False
                        break

            if v.tiny:
                question = '%s (%s)' % (question, '/'.join((s.upper() if s == v.default else s)
                                                            for s in v.choices))
                for s in v.choices:
                    if s == v.default:
                        aliases[s.upper()] = s
                for key, value in v.choices.items():
                    print('     %s%s%s: %s' % (self.BOLD, key, self.NC, value))
            else:
                for n, (key, value) in enumerate(v.choices.items()):
                    print('     %s%2d)%s %s' % (self.BOLD, n + 1, self.NC, value))
                    aliases[str(n + 1)] = key
                question = '%s (choose in list)' % question
        if v.masked:
            question = '%s (hidden input)' % question

        if not isinstance(v, ValueBool) and not v.tiny and v.default not in (None, ''):
            question = '%s [%s]' % (question, '*******' if v.masked else v.default)

        question += ': '

        while True:
            if v.masked:
                line = getpass.getpass(question)
                if sys.platform != 'win32':
                    if isinstance(line, bytes): # only for python2
                        line = line.decode(self.encoding)
            else:
                self.stdout.write(question)
                self.stdout.flush()
                line = self.stdin.readline()
                if len(line) == 0:
                    raise EOFError()
                else:
                    line = line.rstrip('\r\n')

            if not line and v.default is not None:
                line = v.default

            if line in aliases:
                line = aliases[line]

            try:
                v.set(line)
            except ValueError as e:
                print('Error: %s' % e, file=self.stderr)
            else:
                break

        v.noprompt = True
        return v.get()

    def print(self, txt):
        print(txt)

    def acquire_input(
        self,
        content: str | bytes | None = None,
        editor_params: dict[str, str] | None = None,
        suffix: str | None = None
    ) -> str:
        """
        Get an input from a text editor.

        If the EDITOR environment variable is empty, acquire from stdin.

        You can provide ``editor_params`` to supply parameters to editor.
        For example::

            {'vim': "-c 'set ft=mail'"}

        The ``suffix`` argument is for the temporary filename, that's useful to
        highlight content.

        :param content: default content to write in the file to edit
        :type content: str
        :param editor_params: parameters to supply to editor
        :type editor_params: dict
        :param suffix: file suffix
        :type suffix: str
        :rtype: str
        """
        editor = os.getenv('EDITOR', 'vi')
        if self.stdin.isatty() and editor:
            with NamedTemporaryFile(suffix=suffix) as f:
                filename = f.name
                if content is not None:
                    if isinstance(content, str):
                        content = content.encode('utf-8')
                    f.write(content)
                    f.flush()
                try:
                    params = editor_params[os.path.basename(editor)]
                except (KeyError,TypeError):
                    params = ''
                cmd = shlex.split(editor) + shlex.split(params) + [filename]
                subprocess.call(cmd)
                f.seek(0)
                text = f.read()
        else:
            if self.stdin.isatty():
                print('Reading content from stdin... Type ctrl-D '
                          'from an empty line to stop.')
            text = self.stdin.read()
        return to_unicode(text)

    def bcall_error_handler(self, backend, error, backtrace):
        """
        Handler for an exception inside the CallErrors exception.

        This method can be overridden to support more exceptions types.
        """
        if isinstance(error, BrowserQuestion):
            for field in error.fields:
                v = self.ask(field)
                if v:
                    backend.config[field.id].set(v)
        elif isinstance(error, CaptchaQuestion):
            print('Warning(%s): Captcha has been found on login page' % backend.name, file=self.stderr)
        elif isinstance(error, BrowserIncorrectPassword):
            msg = str(error)
            if not msg:
                msg = 'invalid login/password.'
            print('Error(%s): %s' % (backend.name, msg), file=self.stderr)
            if self.ask('Do you want to reconfigure this backend?', default=True):
                self.unload_backends(names=[backend.name])
                self.edit_backend(backend.name)
                self.load_backends(names=[backend.name])
        elif isinstance(error, BrowserSSLError):
            print('FATAL(%s): ' % backend.name + self.BOLD + r'/!\ SERVER CERTIFICATE IS INVALID /!\\' + self.NC, file=self.stderr)
        elif isinstance(error, BrowserHTTPSDowngrade):
            print('FATAL(%s): ' % backend.name + 'Downgrade from HTTPS to HTTP')
        elif isinstance(error, BrowserForbidden):
            msg = str(error)
            print('Error(%s): %s' % (backend.name, msg or 'Forbidden'), file=self.stderr)
        elif isinstance(error, BrowserUnavailable):
            msg = str(error)
            print('Error(%s): %s' % (backend.name, msg or 'Website is unavailable.'), file=self.stderr)
        elif isinstance(error, ActionNeeded):
            msg = str(error)
            print('Error(%s): Action needed on website: %s' % (backend.name, msg), file=self.stderr)
        elif isinstance(error, NotImplementedError):
            print('Error(%s): this feature is not supported yet by this backend.' % backend.name, file=self.stderr)
            print('      %s   To help the maintainer of this backend implement this feature,' % (' ' * len(backend.name)), file=self.stderr)
            print('      %s   please contact us on the project mailing list' % (' ' * len(backend.name)), file=self.stderr)
        elif isinstance(error, NeedInteractiveFor2FA):
            print('Error(%s): You have to run %s in interactive mode to perform a two-factor authentication' % (backend.name, self.APPNAME), file=self.stderr)
        elif isinstance(error, UserError):
            print('Error(%s): %s' % (backend.name, error), file=self.stderr)
        elif isinstance(error, MoreResultsAvailable):
            print('Hint: There are more results for backend %s' % (backend.name), file=self.stderr)
        else:
            print('Bug(%s): %s' % (backend.name, error), file=self.stderr)

            minfo = self.woob.repositories.get_module_info(backend.NAME)
            if minfo and not minfo.is_local():
                if self.options.auto_update:
                    self.woob.repositories.update_repositories(ConsoleProgress(self))

                    # minfo of the new available module
                    minfo = self.woob.repositories.get_module_info(backend.NAME)
                    if (
                        minfo and
                        minfo.version > self.woob.repositories.versions.get(minfo.name) and
                        self.ask(f'A new version of {minfo.name} is available. Do you want to install it?', default=True) and
                        self.install_module(minfo)
                    ):
                        print(f'New version of module {minfo.name} has been installed. Retry to call the command.')
                        return
                else:
                    print('(If --auto-update is passed on the command-line, new versions of the module will be checked automatically)')

            if logging.root.level <= logging.DEBUG:
                print(backtrace, file=self.stderr)
            else:
                return True

    def bcall_errors_handler(self, errors, debugmsg='Use --debug option to print backtraces', ignore=()):
        """
        Handler for the CallErrors exception.
        """
        ask_debug_mode = False
        more_results = set()
        err = 0

        for backend, error, backtrace in errors.errors:
            if isinstance(error, MoreResultsAvailable):
                more_results.add(backend.name)
            elif isinstance(error, ignore):
                continue
            else:
                err = 1
                if self.bcall_error_handler(backend, error, backtrace):
                    ask_debug_mode = True

        if ask_debug_mode:
            print(debugmsg, file=self.stderr)
        elif len(more_results) > 0:
            print('Hint: There are more results available for %s (use option -n or count command)' % (', '.join(more_results)), file=self.stderr)

        return err



# Copyright(C) 2018 Julien Veyssier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import base64
import re

import lxml.etree as ET
import requests
from woob.capabilities.base import empty

__all__ = ['recipe_to_krecipes_xml']


def recipe_to_krecipes_xml(recipe, author=None):
    """
    Export recipe to KRecipes XML string
    """
    sauthor = ''
    if not empty(recipe.author):
        sauthor += '%s@' % recipe.author

    if author is None:
        sauthor += 'Cookboob'
    else:
        sauthor += author

    header = '<?xml version="1.0" encoding="UTF-8" ?>\n'
    initial_xml = '''\
<krecipes version='2.0-beta2' lang='fr' xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' xsi:noNamespaceSchemaLocation='krecipes.xsd'>
<krecipes-recipe id='1'>
</krecipes-recipe>
</krecipes>'''
    doc = ET.fromstring(initial_xml)
    xrecipe = doc.find('krecipes-recipe')
    desc = ET.SubElement(xrecipe, 'krecipes-description')
    title = ET.SubElement(desc, 'title')
    title.text = recipe.title
    authors = ET.SubElement(desc, 'author')
    authors.text = sauthor
    eyield = ET.SubElement(desc, 'yield')
    if not empty(recipe.nb_person):
        amount = ET.SubElement(eyield, 'amount')
        if len(recipe.nb_person) == 1:
            amount.text = '%s' % recipe.nb_person[0]
        else:
            mini = ET.SubElement(amount, 'min')
            mini.text = '%s' % recipe.nb_person[0]
            maxi = ET.SubElement(amount, 'max')
            maxi.text = '%s' % recipe.nb_person[1]
        etype = ET.SubElement(eyield, 'type')
        etype.text = 'persons'
    if not empty(recipe.preparation_time):
        preptime = ET.SubElement(desc, 'preparation-time')
        preptime.text = '%02d:%02d' % (recipe.preparation_time / 60, recipe.preparation_time % 60)
    if recipe.picture and recipe.picture.url:
        data = requests.get(recipe.picture.url, timeout=30).content
        datab64 = base64.encodestring(data)[:-1]

        pictures = ET.SubElement(desc, 'pictures')
        pic = ET.SubElement(pictures, 'pic', {'format': 'JPEG', 'id': '1'})
        pic.text = ET.CDATA(datab64)

    if not empty(recipe.ingredients):
        ings = ET.SubElement(xrecipe, 'krecipes-ingredients')
        pat = re.compile('^[0-9,.]*')
        for i in recipe.ingredients:
            sname = '%s' % i
            samount = ''
            sunit = ''
            first_nums = pat.match(i).group()
            if first_nums != '':
                samount = first_nums
                sname = i.lstrip('0123456789 ')

            ing = ET.SubElement(ings, 'ingredient')
            am = ET.SubElement(ing, 'amount')
            am.text = samount
            unit = ET.SubElement(ing, 'unit')
            unit.text = sunit
            name = ET.SubElement(ing, 'name')
            name.text = sname

    if not empty(recipe.instructions):
        instructions = ET.SubElement(xrecipe, 'krecipes-instructions')
        instructions.text = recipe.instructions

    if not empty(recipe.comments):
        ratings = ET.SubElement(xrecipe, 'krecipes-ratings')
        for c in recipe.comments:
            rating = ET.SubElement(ratings, 'rating')
            if c.author:
                rater = ET.SubElement(rating, 'rater')
                rater.text = c.author
            if c.text:
                com = ET.SubElement(rating, 'comment')
                com.text = c.text
            crits = ET.SubElement(rating, 'criterion')
            if c.rate:
                crit = ET.SubElement(crits, 'criteria')
                critname = ET.SubElement(crit, 'name')
                critname.text = 'Overall'
                critstars = ET.SubElement(crit, 'stars')
                critstars.text = c.rate.split('/')[0]

    return header + ET.tostring(doc, encoding='UTF-8', pretty_print=True).decode('utf-8')




# Copyright(C) 2012 Lucien Loiseau
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, StringField, UserError


__all__ = ['TranslationFail', 'LanguageNotSupported', 'CapTranslate']


class LanguageNotSupported(UserError):
    """
    Raised when the language is not supported
    """

    def __init__(self, msg='language is not supported'):
        super(LanguageNotSupported, self).__init__(msg)


class TranslationFail(UserError):
    """
    Raised when no translation matches the given request
    """

    def __init__(self, msg='No Translation Available'):
        super(TranslationFail, self).__init__(msg)


class Translation(BaseObject):
    """
    Translation.
    """
    lang_src =      StringField('Source language')
    lang_dst =      StringField('Destination language')
    text =          StringField('Translation')


class CapTranslate(Capability):
    """
    Capability of online translation website to translate word or sentence
    """

    def translate(self, source_language, destination_language, request):
        """
        Perfom a translation.

        :param source_language: language in which the request is written
        :param destination_language: language to translate the request into
        :param request: the sentence to be translated
        :rtype: Translation
        """
        raise NotImplementedError()



# Copyright(C) 2010-2012 Romain Bignon, Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import os
from collections import OrderedDict

from woob.capabilities.account import CapAccount
from woob.exceptions import ModuleLoadError
from woob.tools.application.repl import ReplApplication
from woob.tools.application.console import ConsoleProgress
from woob.tools.application.formatters.iformatter import IFormatter

__all__ = ['AppConfig']


class ModuleInfoFormatter(IFormatter):
    def format_dict(self, minfo):
        result = '.------------------------------------------------------------------------------.\n'
        result += '| Module %-69s |\n' % minfo['name']
        result += "+-----------------.------------------------------------------------------------'\n"
        result += '| Version         | %s\n' % minfo['version']
        result += '| Maintainer      | %s\n' % minfo['maintainer']
        result += '| License         | %s\n' % minfo['license']
        result += '| Description     | %s\n' % minfo['description']
        result += '| Capabilities    | %s\n' % ', '.join(minfo['capabilities'])
        result += '| Installed       | %s\n' % minfo['installed']
        result += '| Location        | %s\n' % minfo['location']
        if 'config' in minfo:
            first = True
            for key, field in minfo['config'].items():
                label = field['label']
                if field['default'] is not None:
                    label += ' (default: %s)' % field['default']

                if first:
                    result += '|                 | \n'
                    result += '| Configuration   | %s: %s\n' % (key, label)
                    first = False
                else:
                    result += '|                 | %s: %s\n' % (key, label)
        result += "'-----------------'\n"
        return result


class AppConfig(ReplApplication):
    APPNAME = 'config'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Christophe Benz, Romain Bignon'
    DESCRIPTION = "Console application to add/edit/remove backends, " \
                  "and to register new website accounts."
    SHORT_DESCRIPTION = "manage backends or register new accounts"
    EXTRA_FORMATTERS = {'info_formatter': ModuleInfoFormatter}
    COMMANDS_FORMATTERS = {'modules':     'table',
                           'list':        'table',
                           'info':        'info_formatter',
                           }
    DISABLE_REPL = True

    def load_default_backends(self):
        pass

    def do_add(self, line):
        """
        add MODULE_NAME [BACKEND_NAME] [PARAMETERS ...]

        Create a backend from a module. By default, if BACKEND_NAME is omitted,
        that's the module name which is used.

        You can specify parameters from command line in form "key=value".
        """
        if not line:
            print('You must specify a module name. Hint: use the "modules" command.', file=self.stderr)
            return 2

        module_name, options = self.parse_command_args(line, 2, 1)
        if options:
            options = options.split(' ')
        else:
            options = ()

        backend_name = None

        params = {}
        # set backend params from command-line arguments
        for option in options:
            try:
                key, value = option.split('=', 1)
            except ValueError:
                if backend_name is None:
                    backend_name = option
                else:
                    print('Parameters have to be formatted "key=value"', file=self.stderr)
                    return 2
            else:
                params[key] = value

        self.add_backend(module_name, backend_name or module_name, params)

    def do_register(self, line):
        """
        register MODULE

        Register a new account on a module.
        """
        self.register_backend(line)

    def do_confirm(self, backend_name):
        """
        confirm BACKEND

        For a backend which support CapAccount, parse a confirmation mail
        after using the 'register' command to automatically confirm the
        subscribe.

        It takes mail from stdin. Use it with postfix for example.
        """
        # Do not use the ReplApplication.load_backends() method because we
        # don't want to prompt user to create backend.
        self.woob.load_backends(names=[backend_name])
        try:
            backend = self.woob.get_backend(backend_name)
        except KeyError:
            print('Error: backend "%s" not found.' % backend_name, file=self.stderr)
            return 1

        if not backend.has_caps(CapAccount):
            print('Error: backend "%s" does not support accounts management' % backend_name, file=self.stderr)
            return 1

        mail = self.acquire_input()
        if not backend.confirm_account(mail):
            print('Error: Unable to confirm account creation', file=self.stderr)
            return 1
        return 0

    def do_list(self, line):
        """
        list [CAPS ..]

        Show backends.
        """
        caps = line.split()
        for backend_name, module_name, params in sorted(self.woob.backends_config.iter_backends()):
            try:
                module = self.woob.modules_loader.get_or_load_module(module_name)
            except ModuleLoadError as e:
                self.logger.warning('Unable to load module %r: %s' % (module_name, e))
                continue

            if caps and not module.has_caps(*caps):
                continue
            row = OrderedDict([('Name', backend_name),
                               ('Module', module_name),
                               ('Configuration', ', '.join(
                                   '%s=%s' % (key, ('*****' if key in module.config and module.config[key].masked
                                                    else value))
                                   for key, value in params.items())),
                               ])
            self.format(row)

    def do_enable(self, backend_name):
        """
        enable NAME

        Enable a backend.
        """
        try:
            self.woob.backends_config.edit_backend(backend_name, {'_enabled': '1'})
        except KeyError:
            print('Backend instance "%s" does not exist' % backend_name, file=self.stderr)
            return 1

    def do_disable(self, backend_name):
        """
        disable NAME

        Disable a backend.
        """
        try:
            self.woob.backends_config.edit_backend(backend_name, {'_enabled': '0'})
        except KeyError:
            print('Backend instance "%s" does not exist' % backend_name, file=self.stderr)
            return 1

    def do_remove(self, backend_name):
        """
        remove NAME

        Remove a backend.
        """
        if not self.woob.backends_config.remove_backend(backend_name):
            print('Backend instance "%s" does not exist' % backend_name, file=self.stderr)
            return 1

    def do_edit(self, line):
        """
        edit BACKEND

        Edit a backend
        """
        try:
            self.edit_backend(line)
        except KeyError:
            print('Error: backend "%s" not found' % line, file=self.stderr)
            return 1

    def do_modules(self, line):
        """
        modules [CAPS ...]

        Show available modules.
        """
        caps = line.split()
        for name, info in sorted(self.woob.repositories.get_all_modules_info(caps).items()):
            row = OrderedDict([('Name', name),
                               ('Capabilities', info.capabilities),
                               ('Description', info.description),
                               ('Installed', info.is_installed()),
                               ])
            self.format(row)

    def do_info(self, line):
        """
        info NAME

        Display information about a module.
        """
        if not line:
            print('You must specify a module name. Hint: use the "modules" command.', file=self.stderr)
            return 2

        minfo = self.woob.repositories.get_module_info(line)
        if not minfo:
            print('Module "%s" does not exist.' % line, file=self.stderr)
            return 1

        try:
            module = self.woob.modules_loader.get_or_load_module(line)
        except ModuleLoadError:
            module = None

        self.start_format()
        self.format(self.create_minfo_dict(minfo, module))


    def create_minfo_dict(self, minfo, module):
        module_info = {}
        module_info['name'] = minfo.name
        module_info['version'] = minfo.version
        module_info['maintainer'] = minfo.maintainer
        module_info['license'] = minfo.license
        module_info['description'] = minfo.description
        module_info['capabilities'] = minfo.capabilities
        repo_ver = self.woob.repositories.versions.get(minfo.name)
        module_info['installed'] = '%s%s' % (
            'yes' if module else 'no',
            ' (new version available)' if repo_ver and repo_ver > minfo.version else ''
        )
        module_info['location'] = '%s' % (minfo.url or os.path.join(minfo.path, minfo.name))
        if module:
            module_info['config'] = {}
            for key, field in module.config.items():
                # transient fields are used only for non persistent information
                # and are not provided in the module config.
                if field.transient:
                    continue

                module_info['config'][key] = {'label': field.label,
                                              'default': field.default,
                                              'description': field.description,
                                              'regexp': field.regexp,
                                              'choices': field.choices,
                                              'masked': field.masked,
                                              'required': field.required}
        return module_info

    def do_update(self, line):
        """
        update

        Update woob.
        """
        self.woob.update(ConsoleProgress(self))
        if self.woob.repositories.errors:
            print('Errors building modules: %s' % ', '.join(self.woob.repositories.errors.keys()), file=self.stderr)
            if not self.options.debug:
                print('Use --debug to get more information.', file=self.stderr)
            return 1



# Copyright(C) 2019 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import requests
from urllib3.util.ssl_ import create_urllib3_context


__all__ = ['HTTPAdapter', 'LowSecHTTPAdapter']


class HTTPAdapter(requests.adapters.HTTPAdapter):
    """
    Custom Adapter class with extra features.

    :param proxy_headers: headers to send to proxy (if any)
    :type proxy_headers: dict
    """
    def __init__(self, *args, **kwargs):
        self._proxy_headers = kwargs.pop('proxy_headers', {})
        super().__init__(*args, **kwargs)

    def add_proxy_header(self, key, value):
        self._proxy_headers[key] = value

    def update_proxy_headers(self, headers):
        self._proxy_headers.update(headers)

    def proxy_headers(self, proxy):
        headers = super().proxy_headers(proxy)
        headers.update(self._proxy_headers)
        return headers


class LowSecHTTPAdapter(HTTPAdapter):
    """
    Adapter to use with low security HTTP servers.

    Some websites uses small DH keys, which is deemed insecure by OpenSSL's
    default config.  we have to lower its expectations so it accepts the
    certificate.

    See https://www.ssllabs.com/ssltest/analyze.html?d=www.ibps.bpaura.banquepopulaire.fr
    for the exhaustive list of defects they are too incompetent to fix
    """

    def init_poolmanager(self, *args, **kwargs):
        context = create_urllib3_context(ciphers="DEFAULT:@SECLEVEL=1")
        kwargs['ssl_context'] = context
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        context = create_urllib3_context(ciphers="DEFAULT:@SECLEVEL=1")
        kwargs['ssl_context'] = context
        return super().proxy_manager_for(*args, **kwargs)



# Copyright(C) 2016      Vincent A
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

"""
Module to use NSS instead of OpenSSL in urllib3/requests.

As it was used by Firefox, it offered the same security level, to be sure woob
can access to a website the same way than a real browser.

.. deprecated:: 3.4
   NSS is not maintained by Mozilla anymore, so do not use it anymore.

   If there is an issue to access a website, you can override use the
   :class:`woob.browser.adapters.LowSecHTTPAdapter` with the
   :attr:`woob.browser.browsers.Browser.HTTP_ADAPTER_CLASS` attribute to be
   explicitely more permissive.
"""

# create db:
#   mkdir pki
#   certutil -N -d pki

# import certificate:
#   find -L /etc/ssl/certs -name "*.pem" | while read f; do certutil -A -d pki -i $f -n $f -t TCu,Cu,Tu; done

from functools import wraps
from io import RawIOBase, BufferedRWPair
import hashlib
import os
import re
import socket
import ssl as basessl
import subprocess
from tempfile import NamedTemporaryFile
from threading import Lock
import warnings

try:
    import nss.ssl
    import nss.error
    import nss.nss
except ImportError:
    raise ImportError('Please install python3-nss')
from requests.packages.urllib3.util.ssl_ import ssl_wrap_socket as old_ssl_wrap_socket
import requests  # for AIA
from woob.tools.log import getLogger


warnings.warn('Use of NSS is deprecated, it will be removed in woob 4.0', DeprecationWarning, stacklevel=2)


__all__ = ['init_nss', 'inject_in_urllib3', 'certificate_db_filename']


CTX = None
INIT_PID = None
INIT_ARGS = None
LOGGER = getLogger(__name__)


def nss_version():
    version_str = nss.nss.nss_get_version()
    version_str = re.match(r'\d+\.\d+', version_str).group(0) # can be "3.21.3 Extended ECC"
    return tuple(int(x) for x in version_str.split('.'))
    # see https://developer.mozilla.org/en-US/docs/Mozilla/Projects/NSS/NSS_3.35_release_notes


def certificate_db_filename():
    version = nss_version()
    if version < (3, 35):
        return 'cert8.db'
    return 'cert9.db'


def path_for_version(path):
    # despite what certutil(1) and the NSS 3.35 releases notes say
    # some nss builds >=3.35 will use either sql or dbm by default
    # also, some builds <3.35 will fail to enforce sql format when using "sql:"
    if nss_version() < (3, 35):
        return path
    return 'sql:%s' % path


def cert_to_dict(cert):
    # see https://docs.python.org/2/library/ssl.html#ssl.SSLSocket.getpeercert
    # and https://github.com/kennethreitz/requests/blob/master/requests/packages/urllib3/contrib/pyopenssl.py

    mappings = {
        nss.nss.certDNSName: 'DNS',
        nss.nss.certIPAddress: 'IP Address',
        # TODO support more types
    }

    altnames = []
    try:
        ext = cert.get_extension(nss.nss.SEC_OID_X509_SUBJECT_ALT_NAME)
    except KeyError:
        pass
    else:
        for entry in nss.nss.x509_alt_name(ext.value, nss.nss.AsObject):
            key = mappings[entry.type_enum]
            altnames.append((key, entry.name))

    ret = {
        'subject': [
            [('commonName', cert.subject.common_name)],
            [('localityName', cert.subject.locality_name)],
            [('organizationName', cert.subject.org_name)],
            [('organizationalUnitName', cert.subject.org_unit_name)],
            [('emailAddress', cert.subject.email_address)],
        ],
        'subjectAltName': altnames,
        'issuer': [
            [('countryName', cert.issuer.country_name)],
            [('organizationName', cert.issuer.org_name)],
            [('organizationalUnitName', cert.issuer.org_unit_name)],
            [('commonName', cert.issuer.common_name)],
        ],
        # TODO serialNumber, notBefore, notAfter
        'version': cert.version,
    }

    return ret


ERROR_MAP = {
    nss.error.PR_CONNECT_TIMEOUT_ERROR: (socket.timeout,),
    nss.error.PR_IO_TIMEOUT_ERROR: (socket.timeout,),
    nss.error.PR_CONNECT_RESET_ERROR: (socket.error,),
}


def wrap_callable(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return exc_wrap(func, *args, **kwargs)
    return wrapper


def exc_wrap(func, *args, **kwargs):
    try:
        return func(*args, **kwargs)
    except nss.error.NSPRError as e:
        if e.error_desc.startswith('(SEC_ERROR_') or e.error_desc.startswith('(SSL_ERROR_'):
            raise basessl.SSLError(0, e.error_message or e.error_desc, e)

        for k in ERROR_MAP:
            if k == e.error_code:
                raise ERROR_MAP[k][0]

        raise


class NSSFile(RawIOBase):
    def __init__(self, obj):
        self.obj = obj
        self.open = True

    def close(self):
        super(NSSFile, self).close()
        if self.open:
            self.obj.close()
            self.open = False

    def read(self, amount):
        return self.obj.recv(amount)

    def readinto(self, buf):
        amount = len(buf)
        chunk = self.obj.recv(amount)
        # TODO handle timeout by returning None?
        buf[:len(chunk)] = chunk
        return len(chunk)

    def write(self, buf):
        self.obj.send(buf)
        return len(buf)

    def readable(self):
        return self.open

    writable = readable


class Wrapper:
    def __init__(self, obj):
        self.__obj = obj
        self.__timeout = nss.io.PR_INTERVAL_NO_TIMEOUT

    def settimeout(self, t):
        if t is None:
            self.__timeout = nss.io.PR_INTERVAL_NO_TIMEOUT
        else:
            self.__timeout = nss.io.milliseconds_to_interval(int(t * 1000))

    def __getattr__(self, attr):
        ret = getattr(self.__obj, attr)
        if callable(ret):
            ret = wrap_callable(ret)
        return ret

    def getpeercert(self, binary_form=False):
        # TODO return none or exception in case no cert yet?

        cert = self.__obj.get_peer_certificate()
        if binary_form:
            return cert.der_data
        else:
            return cert_to_dict(cert)

    def makefile(self, *args, **kwargs):
        made = self.__obj.makefile(*args, **kwargs)
        # NSS.io.Socket returns the same object, but increments its internal ref counter
        # close() decreases the counter, and closes if there are no more refs
        # see python NSS source
        assert made is self.__obj

        rw_wrapper = NSSFile(self)
        return BufferedRWPair(rw_wrapper, rw_wrapper)

    def recv(self, amount):
        return exc_wrap(self.__obj.recv, amount, self.__timeout)

    def send(self, s):
        return exc_wrap(self.__obj.send, s, self.__timeout)


def auth_cert_pinning(sock, check_sig, is_server, path):
    cert = sock.get_peer_certificate()

    expected = nss.nss.Certificate(nss.nss.read_der_from_file(path, True))
    return (expected.signed_data.data == cert.signed_data.data)


AIA_CACHE = {}
AIA_LOCK = Lock()


def auth_cert_basic(sock, check_sig, is_server):
    cert = sock.get_certificate()
    db = nss.nss.get_default_certdb()

    # simple case: full cert chain
    try:
        valid = cert.verify_hostname(sock.get_hostname())
    except nss.error.NSPRError:
        return False
    if not valid:
        return False

    required = nss.nss.certificateUsageSSLServer
    try:
        usages = cert.verify_now(db, check_sig, required) & required
    except nss.error.NSPRError:
        return False
    return bool(usages)


def auth_cert_aia_only(sock, check_sig, is_server):
    cert = sock.get_certificate()
    db = nss.nss.get_default_certdb()

    # harder case: the server presents an incomplete cert chain and only has the leaf cert
    # the parent is indicated in the TLS extension called "AIA"

    for ext in cert.extensions:
        if ext.name == 'Authority Information Access':
            aia_text = ext.format()
            aia_text = re.sub(r'\s+', ' ', aia_text)
            break
    else:
        return False
    # yes, the parent TLS cert is behind an HTTP URL
    parent_url = re.search(r'Method: PKIX CA issuers access method Location: URI: (http:\S+)', aia_text).group(1)

    with AIA_LOCK:
        parent_der = AIA_CACHE.get(parent_url)

    if parent_der is None:
        parent_der = requests.get(parent_url, timeout=30).content
        with AIA_LOCK:
            AIA_CACHE[parent_url] = parent_der

    # verify parent cert is a CA in our db
    parent = nss.nss.Certificate(parent_der, perm=False)
    required = nss.nss.certificateUsageAnyCA
    try:
        usages = parent.verify_now(db, check_sig, required) & required
    except nss.error.NSPRError:
        return False
    if not usages:
        return False

    # verify leaf certificate
    try:
        valid = cert.verify_hostname(sock.get_hostname())
    except nss.error.NSPRError:
        return False
    if not valid:
        return False

    required = nss.nss.certificateUsageSSLServer
    try:
        usages = cert.verify_now(db, check_sig, required) & required
    except nss.error.NSPRError:
        return False
    return bool(usages)


def auth_cert_with_aia(sock, check_sig, is_server):
    assert not is_server

    cert = sock.get_certificate()

    if len(cert.get_cert_chain()) > 1:
        return auth_cert_basic(sock, check_sig, is_server)
    else:
        return auth_cert_aia_only(sock, check_sig, is_server)


DEFAULT_CA_CERTIFICATES = (
    '/etc/ssl/certs/ca-certificates.crt',
    '/etc/pki/tls/certs/ca-bundle.crt',
)

try:
    import certifi
except ImportError:
    pass
else:
    DEFAULT_CA_CERTIFICATES = DEFAULT_CA_CERTIFICATES + (certifi.where(),)


def ssl_wrap_socket(sock, *args, **kwargs):
    if kwargs.get('certfile'):
        LOGGER.debug('a client certificate is used, falling back to OpenSSL')
        # TODO implement NSS client certificate support
        return old_ssl_wrap_socket(sock, *args, **kwargs)

    reinit_if_needed()

    # TODO handle more options?
    hostname = kwargs.get('server_hostname')
    ossl_ctx = kwargs.get('ssl_context')

    # the python Socket and the NSS SSLSocket are agnostic of each other's state
    # so the Socket could close the fd, then a file could be opened,
    # obtaining the same file descriptor, then NSS would use the file, thinking
    # it's a network file descriptor... dup the fd to make it independant
    fileno = sock.fileno()
    if hasattr(sock, 'detach'):
        # socket.detach only exists in py3.
        sock.detach()
    else:
        fileno = os.dup(fileno)

    nsssock = nss.ssl.SSLSocket.import_tcp_socket(fileno)
    wrapper = Wrapper(nsssock)

    nsssock.set_certificate_db(nss.nss.get_default_certdb())
    if hostname:
        nsssock.set_hostname(hostname)
    if ossl_ctx and not ossl_ctx.verify_mode:
        nsssock.set_auth_certificate_callback(lambda *args: True)
    elif kwargs.get('ca_certs') and kwargs['ca_certs'] not in DEFAULT_CA_CERTIFICATES:
        nsssock.set_auth_certificate_callback(auth_cert_pinning, kwargs['ca_certs'])
    else:
        nsssock.set_auth_certificate_callback(auth_cert_with_aia)

    nsssock.reset_handshake(False) # marks handshake as not-done
    try:
        wrapper.send(b'') # performs handshake
    except nss.error.NSPRError as e:
        if e.error_code == nss.error.PR_END_OF_FILE_ERROR:
            # the corresponding openssl error isn't exactly socket.timeout()
            # but rather something SyscallError.
            # i don't know how to generate it exactly and the end result is
            # similar so let's use this.
            raise socket.timeout()

        # see below why closing
        wrapper.close()
        raise
    except:
        # If there is an exception during the handshake, correctly close the
        # duplicated/detached socket as it isn't known by the caller.
        wrapper.close()
        raise

    return wrapper


def inject_in_urllib3():
    import urllib3.util.ssl_
    import urllib3.connection
    # on some distros, requests comes with its own urllib3 version
    import requests.packages.urllib3.util.ssl_
    import requests.packages.urllib3.connection

    for pkg in (urllib3, requests.packages.urllib3):
        pkg.util.ssl_.ssl_wrap_socket = ssl_wrap_socket
        pkg.util.ssl_wrap_socket = ssl_wrap_socket
        pkg.connection.ssl_wrap_socket = ssl_wrap_socket


def init_nss(path, rw=False):
    global CTX, INIT_PID, INIT_ARGS

    if CTX is not None and INIT_PID == os.getpid():
        return

    INIT_ARGS = (path, rw)

    if rw:
        flags = 0
    else:
        flags = nss.nss.NSS_INIT_READONLY

    path = path_for_version(path)
    INIT_PID = os.getpid()
    CTX = nss.nss.nss_init_context(path, flags=flags)


def add_nss_cert(dbpath, certpath, nickname):
    # Even if you use a different nickname, NSS will not add a cert that is
    # already in db, without signaling it.
    subprocess.check_call(['certutil', '-A', '-d', dbpath, '-i', certpath, '-n', nickname, '-t', 'TC,C,T'])


def del_nss_cert(dbpath, nickname):
    subprocess.check_call(['certutil', '-D', '-d', dbpath, '-n', nickname])


def create_cert_db(path):
    # continue to provide this function for braindead customers who believe a development version
    # is an api-stable version
    update_cert_db(path)


def iter_db_certs(path):
    # TODO check existing db
    output = subprocess.check_output(['certutil', '-L', '-d', path]).decode('utf-8').rstrip()
    for line in output.split('\n'):
        if line.startswith(' ') or ',' not in line:
            # NSS prints a useless header.
            # The lines we want have format:
            #   {cert nickname}{space based alignment}{nss trust flags}
            # NSS trust flags contain "," so we're guaranteed these are the desired lines
            continue

        yield line.split()[0]


def create_empty_db(path):
    path = path_for_version(path)

    try:
        subprocess.check_call(['certutil', '-N', '--empty-password', '-d', path])
    except OSError:
        raise ImportError('Please install libnss3-tools')


def update_cert_db(dbpath):
    """Imports certificates from system dir into NSS database."""

    # Tries to keep unchanged certificates.
    # Each certificate has a "nickname" in NSS db, which is defined by us, not
    # by cert content.
    # Previously, we used the certificate file path as nickname.
    # To be able to track those which changed and those removed, we use the
    # hash of the PEM data as nickname. This is useful for ensuring unicity as
    # system certificate dirs contain duplicate certificates.

    realdbpath = dbpath
    dbpath = path_for_version(dbpath)

    if not os.path.exists(os.path.join(realdbpath, certificate_db_filename())):
        create_empty_db(realdbpath)

    db_certs = set(iter_db_certs(dbpath))
    obsolete_certs = set(db_certs)

    # Do a first pass by removing certs that use file nicknames.
    # NSS won't add duplicate certs, so existing file nicknames will prevent
    # hash nicknames from being inserted into the db.
    # Then, the purging loop at function end will remove file nicknames, which
    # will make the db effectively empty, because we couldn't insert any hash
    # nicknames...
    # If we clean them before, we aren't blocked from adding hash nicknames.
    for nick in db_certs:
        if '/' in nick:
            del_nss_cert(dbpath, nick)
            obsolete_certs.discard(nick)

    cert_dir = '/etc/ssl/certs'
    for cert_file in os.listdir(cert_dir):
        cert_file = os.path.join(cert_dir, cert_file)
        if os.path.isdir(cert_file) or '.' not in cert_file:
            continue

        with open(cert_file) as fd:
            content = fd.read()

        separators = [
            '-----END CERTIFICATE-----',
            '-----END TRUSTED CERTIFICATE-----',
        ]
        for sep in separators:
            if sep in content:
                separator = sep
                break
        else:
            continue

        nb_certs = content.count(separator)
        try:
            if nb_certs == 1:
                nick = hashlib.sha1(content.strip().encode('utf-8')).hexdigest()
                if nick in db_certs:  # no need to import it
                    obsolete_certs.discard(nick)  # don't remove it at the end
                    continue

                add_nss_cert(dbpath, cert_file, nick)
                db_certs.add(nick)
            elif nb_certs > 1:
                # nss can't import bundles, split them
                for subcert in content.split(separator)[:-1]:
                    subcert += separator

                    nick = hashlib.sha1(subcert.strip().encode('utf-8')).hexdigest()
                    if nick in db_certs:
                        obsolete_certs.discard(nick)
                        continue

                    with NamedTemporaryFile('w') as fd:
                        fd.write(subcert)
                        fd.flush()
                        add_nss_cert(dbpath, fd.name, nick)
                        db_certs.add(nick)

        except subprocess.CalledProcessError:
            LOGGER.warning('Unable to handle ca file {}'.format(cert_file))

    for nick in obsolete_certs:
        # Those certs were imported in a previous session, but they don't seem
        # to be on the system anymore.
        del_nss_cert(dbpath, nick)


def reinit_if_needed():
    # if we forked since NSS was initialized, we might get an exception
    # (SEC_ERROR_PKCS11_DEVICE_ERROR) A PKCS #11 module returned CKR_DEVICE_ERROR, indicating that a problem has occurred with the token or slot.
    # so we should reinit NSS

    if INIT_PID and INIT_PID != os.getpid():
        LOGGER.debug('nss inited in %s but now in %s', INIT_PID, os.getpid())
        assert INIT_ARGS
        init_nss(*INIT_ARGS)



# Copyright(C) 2018 Quentin Defenouillere
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.capabilities.base import empty
from woob.capabilities.bands import CapBands
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import PrettyFormatter


__all__ = ['Appbands', 'BandInfoFormatter', 'BandListFormatter', 'FavoritesFormatter']


class BandInfoFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'genre', 'year', 'country', 'description')
    def format_obj(self, obj, alias):
        result = '\n%s%s%s\n' % (self.BOLD, obj.name, self.NC)
        if not empty(obj.genre):
            result += 'Genre: %s\n' % obj.genre
        if not empty(obj.country):
            result += 'Country: %s\n' % obj.country
        if not empty(obj.year):
            result += 'Formed in: %s\n' % obj.year
        if not empty(obj.description):
            result += '%sDescription:%s\n' % (self.BOLD, self.NC)
            result += '%s\n' % obj.description
        return result.strip()


class BandListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'short_description')

    def get_title(self, obj):
        return obj.name

    def get_description(self, obj):
        result = ''
        if not empty(obj.short_description):
            result += '%s\n' % obj.short_description
        result+='---------------------------------------------------------------------------'
        return result.strip()


class FavoritesFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'band_url', 'short_description')

    def get_title(self, obj):
        return obj.name

    def get_description(self, obj):
        result = ''
        if not empty(obj.short_description):
            result += '%s\n' % obj.short_description
        if not empty(obj.band_url):
            result += '\t%s\n' % obj.band_url
        result+='---------------------------------------------------------------------------'
        return result.strip()


class AlbumsFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'album_type', 'year', 'reviews')

    def format_obj(self, obj, alias):
        result = '\n%s%s%s\n' % (self.BOLD, obj.name, self.NC)
        if not empty(obj.album_type):
            result += 'Album type: %s\n' % obj.album_type
        if not empty(obj.id):
            result += 'Link to album: %s\n' % obj.id
        if not empty(obj.year):
            result += 'Year of release: %s\n' % obj.year
        if not empty(obj.reviews):
            result += 'Reviews: %s\n' % obj.reviews
        result+='---------------------------------------------------------------------------'
        return result.strip()


class SuggestionsFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'description', 'url')

    def get_title(self, obj):
        return obj.name

    def get_description(self, obj):
        result = ''
        if not empty(obj.description):
            result += '%s\n' % obj.description
        if not empty(obj.url):
            result += '\tLink to band: %s\n' % obj.url
        result+='---------------------------------------------------------------------------'
        return result.strip()



class Appbands(ReplApplication):
    APPNAME = 'bands'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2018-YEAR Quentin Defenouillere'
    DESCRIPTION = "Console application allowing to display music bands and offer music suggestions."
    SHORT_DESCRIPTION = "display bands and suggestions"
    CAPS = CapBands
    DEFAULT_FORMATTER = 'table'
    EXTRA_FORMATTERS = {
        'band_search': BandListFormatter,
        'band_info': BandInfoFormatter,
        'get_favorites': FavoritesFormatter,
        'get_albums': AlbumsFormatter,
        'suggestion': SuggestionsFormatter
    }

    COMMANDS_FORMATTERS = {
        'search': 'band_search',
        'info': 'band_info',
        'favorites': 'get_favorites',
        'albums': 'get_albums',
        'suggestions': 'suggestion'
    }

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    @defaultcount(20)
    def do_search(self, pattern):
        """
        band PATTERN
        Search bands.
        """
        self.change_path(['search'])
        self.start_format()
        for band in self.do('iter_band_search', pattern, caps=CapBands):
            self.cached_format(band)

    def do_info(self, line):
        """
        info BAND_ID
        Get detailed info for specified band. Use the 'search' command to find bands.
        """
        band, = self.parse_command_args(line, 1, 1)
        _id, backend_name = self.parse_id(band)

        self.start_format()
        for info in self.do('get_info', _id, backends=backend_name, caps=CapBands):
            if info:
                self.format(info)

    @defaultcount(40)
    def do_albums(self, line):
        """
        albums BAND_ID
        Get the discography of a band.
        """
        albums, = self.parse_command_args(line, 1, 1)
        _id, backend_name = self.parse_id(albums)

        self.start_format()
        for album in self.do('get_albums', _id, backends=backend_name, caps=CapBands):
            self.cached_format(album)

    @defaultcount(100)
    def do_favorites(self, *ignored):
        """
        favorites
        Displays your favorite bands.
        """
        self.change_path(['favorites'])
        self.start_format()
        for favorite in self.do('get_favorites', caps=CapBands):
            self.cached_format(favorite)

    @defaultcount(100)
    def do_suggestions(self, *ignored):
        """
        suggestions
        Suggests bands depending on your favorite bands.
        """
        self.change_path(['suggestions'])
        self.start_format()
        for suggestion in self.do('suggestions', caps=CapBands):
            self.cached_format(suggestion)



# Copyright(C) 2021 woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.base import empty, NotLoaded
from woob.capabilities.messages import Thread


def _first_non_empty(*args, default=NotLoaded):
    for arg in args:
        if not empty(arg):
            return arg
    return default


def build_linear_thread(messages, thread=None, title=None):
    if thread is None:
        thread = Thread()
        thread.flags = Thread.IS_DISCUSSION

    it = iter(messages)

    root = next(it)
    root.thread = thread
    thread.root = root

    thread.title = _first_non_empty(thread.title, root.title, title)
    root.title = _first_non_empty(root.title, thread.title, title)

    # empty string by default, can't use _first_non_empty
    thread.id = thread.id or root.id or NotLoaded

    thread.url = thread.url or root.url

    if empty(thread.date):
        thread.date = root.date

    root.children = []

    for message in it:
        message.thread = thread
        if empty(message.title) and not empty(root.title):
            message.title = f"Re: {root.title}"
        root.children.append(message)

    return thread



# Copyright(C) 2012  Lucien Loiseau
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import re

from woob.capabilities.translate import CapTranslate, TranslationFail, LanguageNotSupported
from woob.tools.application.repl import ReplApplication
from woob.tools.application.formatters.iformatter import IFormatter
from babel.core import Locale, UnknownLocaleError
from babel.localedata import locale_identifiers

__all__ = ['AppTranslate']


class TranslationFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'text')

    def format_obj(self, obj, alias):
        return '%s* %s%s\n\t[%s] %s' % (self.BOLD, obj.backend, self.NC, obj.lang_dst, obj.text.replace('\n', '\n\t'))


class XmlTranslationFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'text')

    def start_format(self, **kwargs):
        if 'source' in kwargs:
            self.output('<source>\n%s\n</source>' % kwargs['source'])

    def format_obj(self, obj, alias):
        return '<translation %s>\n%s\n</translation>' % (obj.backend, obj.text)


class AppTranslate(ReplApplication):
    APPNAME = 'translate'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2012-YEAR Lucien Loiseau'
    DESCRIPTION = "Console application to translate text from one language to another"
    SHORT_DESCRIPTION = "translate text from one language to another"
    CAPS = CapTranslate
    EXTRA_FORMATTERS = {'translation': TranslationFormatter,
                        'xmltrans':    XmlTranslationFormatter}
    COMMANDS_FORMATTERS = {'translate': 'translation'}

    def parse_lang(self, s):
        try:
            locale = Locale.parse(s.replace('-', '_'))
        except UnknownLocaleError:
            pattern = re.compile(r'\b%s\b' % re.escape(s), re.I)
            for locale_id in locale_identifiers():
                locale = Locale.parse(locale_id)
                if pattern.search(locale.english_name):
                    return locale.language
            return s
        else:
            return locale.language

    def do_translate(self, line):
        """
        translate FROM TO [TEXT]

        Translate from one language to another.
        * FROM : source language
        * TO   : destination language
        * TEXT : text to translate. If "-" is passed, spawn $EDITOR (if in a TTY) to edit text
                 or use standard input text.

        Language  Abbreviation
        ----------------------
        Arabic      ar          Esperanto   eo          Irish       ga          Russian     ru
        Afrikaans   af          Estonian    et          Italian     it          Serbian     sr
        Albanian    sq          Filipino    tl          Japanese    ja          Slovak      sk
        Armenian    hy          Finnish     fi          Kannada     kn          Slovenian   sl
        Azerbaijani az          French      fr          Korean      ko          Spanish     es
        Basque      eu          Galician    gl          Latin       la          Swahili     sw
        Belarusian  be          Georgian    ka          Latvian     lv          Swedish     sv
        Bengali     bn          German      de          Lithuanian  lt          Tamil       ta
        Bulgarian   bg          Greek       gr          Macedonian  mk          Telugu      te
        Catalan     ca          Gujarati    gu          Malay       ms          Thai        th
        Chinese     zh          Haitian     ht          Maltese     mt          Turkish     tr
        Croatian    hr          Hebrew      iw          Norwegian   no          Ukrainian   uk
        Czech       cz          Hindi       hi          Persian     fa          Urdu        ur
        Danish      da          Hungaric    hu          Polish      pl          Vietnamese  vi
        Dutch       nl          Icelandic   is          Portuguese  pt          Welsh       cy
        English     en          Indonesian  id          Romanian    ro          Yiddish     yi
        ----------------------
        """

        lan_from, lan_to, text = self.parse_command_args(line, 3, 2)

        try:
            lan_from = self.parse_lang(lan_from)
            lan_to = self.parse_lang(lan_to)

            if not text or text == '-':
                text = self.acquire_input()

            self.start_format(source=text)
            for translation in self.do('translate', lan_from, lan_to, text):
                self.format(translation)
        except (TranslationFail, LanguageNotSupported) as error:
            print(error, file=self.stderr)
            pass



# Copyright(C) 2014  Simon Murail
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import re
from ast import literal_eval

from woob.browser.filters.standard import Filter, Regexp, RegexpError, FormatError, ItemNotFound


__all__ = ['JSPayload', 'JSValue', 'JSVar']


def _quoted(q):
    return r'(?<!\\){0}(?:\\{0}|[^{0}])*{0}'.format(q)


class JSPayload(Filter):
    r"""
    Get Javascript code from tag's text, cleaned from all comments.

    It filters code in a such a way that corner cases are handled, such as
    comments in string literals and comments in comments.

    The following snippet is borrowed from <https://ostermiller.org/findcomment.html>:

    >>> JSPayload.filter('''someString = "An example comment: /* example */";
    ...
    ... // The comment around this code has been commented out.
    ... // /*
    ... some_code();
    ... // */''')
    'someString = "An example comment: /* example */";\n\nsome_code();\n'

    """
    _single_line_comment = r'[ \t\v\f]*//.*\r?(?:\n|$)'
    _multi_line_comment = r'/\*(?:.|[\r\n])*?\*/'
    _splitter = re.compile('(?:(%s|%s)|%s|%s)' % (_quoted('"'),
                                                  _quoted("'"),
                                                  _single_line_comment,
                                                  _multi_line_comment))

    @classmethod
    def filter(cls, value):
        return ''.join(filter(bool, cls._splitter.split(value)))


class JSValue(Regexp):
    r"""
    Get one or many JavaScript literals.

    It only understands literal values, but should parse them well. Values
    are converted in python values, quotes and slashes in strings are stripped.

    >>> JSValue().filter('boringVar = "boring string"')
    u'boring string'
    >>> JSValue().filter('somecode(); doConfuse(0xdead, cat);')
    57005
    >>> JSValue(need_type=int, nth=2).filter('fazboo("3", "5", 7, "9");')
    7
    >>> JSValue(nth='*').filter('foo([1, 2, 3], "blah", 5.0, true, null]);')
    [1, 2, 3, u'blah', 5.0, True, None]
    """
    pattern = r"""(?x:                                # re.X
        (?:(?P<float>(?:[-+]\s*)?                     # float ?
               (?:(?:\d+\.\d*|\d*\.\d+)(?:[eE]\d+)?
                 |\d+[eE]\d+))
          |(?P<int>(?:[-+]\s*)?(?:0[bB][01]+          # int ?
                                 |0[oO][0-7]+
                                 |0[xX][0-9a-fA-F]+
                                 |\d+))
          |(?:(?:(?:new\s+)?String\()?(?P<str>(?:%s|%s)))  # str ?
          |(?P<bool>true|false)                       # bool ?
          |(?P<None>null))                            # None ?
    )""" % (_quoted('"'), _quoted("'"))

    def to_python(self, m):
        "Convert MatchObject to python value"
        values = m.groupdict()
        for t, v in values.items():
            if v is not None:
                break
        if self.need_type and t != self.need_type:
            raise ItemNotFound('Value with type %s not found' % self.need_type)
        if t in ('int', 'float'):
            return literal_eval(v)
        if t == 'str':
            return literal_eval(v)
        if t == 'bool':
            return v == 'true'
        if t == 'None':
            return
        if self.default:
            return self.default
        raise FormatError('Unable to parse %r value' % m.group(0))

    def __init__(self, selector=None, need_type=None, **kwargs):
        assert 'pattern' not in kwargs and 'flags' not in kwargs, \
               "It would be meaningless to define a pattern and/or flags, use Regexp"
        assert 'template' not in kwargs, "Can't use a template, use Regexp if you have to"
        self.need_type = need_type.__name__ if type(need_type) == type else need_type
        super(JSValue, self).__init__(selector, pattern=self.pattern, template=self.to_python, **kwargs)


class JSVar(JSValue):
    r"""
    Get assigned value of a variable, either as an initialisation value, either
    as an assignement. One can use Regexp's nth parameter to be more specific.

    See JSValue for more details about parsed values.

    >>> JSVar(var='test').filter("var test = .1;\nsomecode()")
    0.1
    >>> JSVar(var='test').filter("test = 666;\nsomecode()")
    666
    >>> JSVar(var='test').filter("test = 'Some \\'string\\' value, isn\\'t it ?';\nsomecode()")
    u"Some 'string' value, isn't it ?"
    >>> JSVar(var='test').filter('test = "Some \\"string\\" value";\nsomecode()')
    u'Some "string" value'
    >>> JSVar(var='test').filter("var test = false;\nsomecode()")
    False
    >>> JSVar(var='test', nth=1).filter("var test = false; test = true;\nsomecode()")
    True
    """
    pattern_template = r"""(?x:                       # re.X
        (?:var\s+)?                                   # optional var keyword
        \b%s                                          # var name
        \s*=\s*                                       # equal sign
    )""" + JSValue.pattern

    def __init__(self, selector=None, var=None, need_type=None, **kwargs):
        assert var is not None, 'Please give a var parameter'
        self.var = var
        self.pattern = self.pattern_template % re.escape(var)
        super(JSVar, self).__init__(selector, need_type=need_type, **kwargs)

    def filter(self, txt):
        try:
            return super(JSVar, self).filter(txt)
        except RegexpError:
            raise ItemNotFound('Variable %r not found' % self.var)



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import codecs
import os
import stat
import sys
from collections.abc import MutableMapping
from configparser import RawConfigParser, DuplicateSectionError
from logging import warning
from subprocess import check_output, CalledProcessError
from typing import Iterator, Tuple


__all__ = ['BackendsConfig', 'BackendAlreadyExists']


class BackendAlreadyExists(Exception):
    """
    Try to add a backend that already exists.
    """


class DictWithCommands(MutableMapping):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._raw = dict(*args, **kwargs)

    def __getitem__(self, key):
        value = self._raw[key]
        if value.startswith('`') and value.endswith('`'):
            try:
                value = check_output(value[1:-1], shell=True)  # nosec: this is intended
            except CalledProcessError as e:
                raise ValueError(f'The call to the external tool failed: {e}') from e

            value = value.decode('utf-8').partition('\n')[0].strip('\r\n\t')

        return value

    def __setitem__(self, key, value):
        self._raw[key] = value

    def __delitem__(self, key):
        del self._raw[key]

    def __len__(self):
        return len(self._raw)

    def __iter__(self):
        return iter(self._raw)


class BackendsConfig:
    """
    Config of backends.

    A backend is an instance of a module with a config.
    A module can therefore have multiple backend instances.

    :param confpath: path to the backends config file
    :type confpath: str
    """

    class WrongPermissions(Exception):
        """
        Unable to write in the backends config file.
        """

    def __init__(self, confpath: str):
        self.confpath = confpath
        try:
            mode = os.stat(confpath).st_mode
        except OSError:
            if not os.path.isdir(os.path.dirname(confpath)):
                os.makedirs(os.path.dirname(confpath))
            if sys.platform == 'win32':
                fptr = open(confpath, 'w', encoding='utf-8')
                fptr.close()
            else:
                try:
                    fd = os.open(confpath, os.O_WRONLY | os.O_CREAT, 0o600)
                    os.close(fd)
                except OSError:
                    fptr = open(confpath, 'w', encoding='utf-8')
                    fptr.close()
                    os.chmod(confpath, 0o600)
        else:
            if sys.platform != 'win32':
                if mode & stat.S_IRGRP or mode & stat.S_IROTH:
                    raise self.WrongPermissions(
                        f'Woob will not start as long as config file {confpath} is readable by group or other users.'
                    )

    def _read_config(self):
        config = RawConfigParser()
        with codecs.open(self.confpath, 'r', encoding='utf-8') as fd:
            config.read_file(fd, self.confpath)
        return config

    def _write_config(self, config):
        f = codecs.open(self.confpath, 'wb', encoding='utf-8')
        with f:
            config.write(f)

    def iter_backends(self) -> Iterator[Tuple[str, str, DictWithCommands]]:
        """
        Iter on all saved backends.

        An item is a tuple with backend name, module name, and params dict.
        """
        config = self._read_config()
        changed = False
        for backend_name in config.sections():
            params = DictWithCommands(config.items(backend_name))
            try:
                module_name = params.pop('_module')
            except KeyError:
                warning('Missing field "_module" for configured backend "%s"', backend_name)
                continue
            yield backend_name, module_name, params

        if changed:
            self._write_config(config)

    def backend_exists(self, name: str) -> bool:
        """
        Return True if the backend exists in config.
        """
        config = self._read_config()
        return name in config.sections()

    def add_backend(self, backend_name: str, module_name: str, params: dict):
        """
        Add a backend to config.

        :param backend_name: name of the backend in config
        :type backend_name: str
        :param module_name: name of woob module
        :type module_name: str
        :param params: params of the backend
        :type params: dict
        """
        if not backend_name:
            raise ValueError('Please give a name to the configured backend.')
        config = self._read_config()
        try:
            config.add_section(backend_name)
        except DuplicateSectionError as exc:
            raise BackendAlreadyExists(backend_name) from exc

        config.set(backend_name, '_module', module_name)
        for key, value in params.items():
            config.set(backend_name, key, value)

        self._write_config(config)

    def edit_backend(self, backend_name: str, params: dict):
        """
        Edit a backend in config.

        :param backend_name: name of the backend in config
        :param params: params to change
        :type params: :class:`dict`
        """
        config = self._read_config()
        if not config.has_section(backend_name):
            raise KeyError(f'Configured backend "{backend_name}" not found')

        for key, value in params.items():
            config.set(backend_name, key, value)

        self._write_config(config)

    def get_backend(self, backend_name: str) -> Tuple[str, dict]:
        """
        Get options of backend.

        :returns: a tuple with the module name and the backends params
        :rtype: tuple[str, dict]
        """

        config = self._read_config()
        if not config.has_section(backend_name):
            raise KeyError(f'Configured backend "{backend_name}" not found')

        # XXX why not a DictWithCommands?
        items = dict(config.items(backend_name))

        try:
            module_name = items.pop('_module')
        except KeyError:
            warning('Missing field "_module" for configured backend "%s"', backend_name)
            raise KeyError(f'Configured backend "{backend_name}" not found')

        return module_name, items

    def remove_backend(self, backend_name: str) -> bool:
        """
        Remove a backend from config.

        Returns False if the backend does not exist.
        """
        config = self._read_config()
        if not config.remove_section(backend_name):
            return False
        self._write_config(config)
        return True



# Copyright(C) 2011-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from base64 import b64decode, b64encode
import os
import codecs
import re
from random import choice

from woob.capabilities.paste import CapPaste, PasteNotFound
from woob.tools.application.repl import ReplApplication


__all__ = ['AppPaste']


class AppPaste(ReplApplication):
    APPNAME = 'paste'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2011-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to post and get pastes from pastebins."
    SHORT_DESCRIPTION = "post and get pastes from pastebins"
    CAPS = CapPaste

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    def do_info(self, line):
        """
        info ID [ID2 [...]]

        Get information about pastes.
        """
        if not line:
            print('This command takes an argument: %s' % self.get_command_help('info', short=True), file=self.stderr)
            return 2

        self.start_format()
        for _id in line.split(' '):
            paste = self.get_object(_id, 'get_paste', ['id', 'title', 'language', 'public', 'contents'])
            if not paste:
                print('Paste not found: %s' % _id, file=self.stderr)

            self.format(paste)


    def do_get(self, line):
        """
        get ID

        Get a paste contents.
        """
        return self._get_op(line, binary=False, command='get')

    def do_get_bin(self, line):
        """
        get_bin ID

        Get a paste contents.
        File will be downloaded from binary services.
        """
        return self._get_op(line, binary=True, command='get_bin')

    def _get_op(self, _id, binary, command='get'):
        if not _id:
            print('This command takes an argument: %s' % self.get_command_help(command, short=True), file=self.stderr)
            return 2

        try:
            paste = self.get_object(_id, 'get_paste', ['contents'])
        except PasteNotFound:
            print('Paste not found: %s' % _id, file=self.stderr)
            return 3
        if not paste:
            print('Unable to handle paste: %s' % _id, file=self.stderr)
            return 1

        if binary:
            if self.interactive:
                if not self.ask('The console may become messed up. Are you sure you want to show a binary file on your terminal?', default=False):
                    print('Aborting.', file=self.stderr)
                    return 1
            output = self.stdout.buffer
            output.write(b64decode(paste.contents))
        else:
            output = self.stdout
            output.write(paste.contents)
            # add a newline unless we are writing
            # in a file or in a pipe
            if output.isatty():
                output.write('\n')

    def do_post(self, line):
        """
        post [FILENAME]

        Submit a new paste.
        The filename can be '-' for reading standard input (pipe).
        If 'bin' is passed, file will be uploaded to binary services.
        """
        return self._post(line, binary=False)

    def do_post_bin(self, line):
        """
        post_bin [FILENAME]

        Submit a new paste.
        The filename can be '-' for reading standard input (pipe).
        File will be uploaded to binary services.
        """
        return self._post(line, binary=True)

    def _post(self, filename, binary):
        use_stdin = (not filename or filename == '-')
        if use_stdin:
            if binary:
                contents = self.stdin.buffer.read()
            else:
                contents = self.acquire_input()
            if not len(contents):
                print('Empty paste, aborting.', file=self.stderr)
                return 1

        else:
            try:
                if binary:
                    m = open(filename, 'rb')
                else:
                    m = codecs.open(filename, encoding=self.options.encoding or self.encoding)
                with m as fp:
                    contents = fp.read()
            except IOError as e:
                print('Unable to open file "%s": %s' % (filename, e.strerror), file=self.stderr)
                return 1

        if binary:
            contents = b64encode(contents).decode('ascii')

        # get and sort the backends able to satisfy our requirements
        params = self.get_params()
        backends = {}
        for backend in self.woob.iter_backends():
            score = backend.can_post(contents, **params)
            if score:
                backends.setdefault(score, []).append(backend)
        # select a random backend from the best scores
        if len(backends):
            backend = choice(backends[max(list(backends.keys()))])
        else:
            print('No suitable backend found.', file=self.stderr)
            return 1

        p = backend.new_paste(_id=None)
        p.public = params['public']
        if self.options.title is not None:
            p.title = self.options.title
        else:
            p.title = os.path.basename(filename)
        p.contents = contents
        backend.post_paste(p, max_age=params['max_age'])
        print('Successfuly posted paste: %s' % p.page_url)

    def get_params(self):
        return {'public': self.options.public,
                'max_age': self.str_to_duration(self.options.max_age),
                'title': self.options.title}

    def str_to_duration(self, s):
        if s.strip().lower() == 'never':
            return False

        parts = re.findall(r'(\d*(?:\.\d+)?)\s*([A-z]+)', s)
        argsmap = {'Y|y|year|years|yr|yrs': 365.25 * 24 * 3600,
                   'M|o|month|months': 30.5 * 24 * 3600,
                   'W|w|week|weeks': 7 * 24 * 3600,
                   'D|d|day|days': 24 * 3600,
                   'H|h|hours|hour|hr|hrs': 3600,
                   'm|i|minute|minutes|min|mins': 60,
                   'S|s|second|seconds|sec|secs': 1}

        seconds = 0
        for number, unit in parts:
            for rx, secs in argsmap.items():
                if re.match('^(%s)$' % rx, unit):
                    seconds += float(number) * float(secs)

        return int(seconds)

    def add_application_options(self, group):
        group.add_option('-p', '--public',  action='store_true',
                         help='Make paste public.')
        group.add_option('-t', '--title', action='store',
                         help='Paste title',
                         type='string')
        group.add_option('-m', '--max-age', action='store',
                         help='Maximum age (duration), default "1 month", "never" for infinite',
                         type='string', default='1 month')
        group.add_option('-E', '--encoding', action='store',
                         help='Input encoding',
                         type='string')



BANK_TRANSACTION_CODES = {
    # Defines every combination for bank transaction codes.
    # This allows to check if the code returned by the bank
    # is valid according to the ISO20022 standard.
    'ACMT': {
        'ACOP': (
            'PSTE',
            'BCKV',
            'ERTA',
            'FLTA',
            'VALD',
            'YTDA',
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'ADOP': (
            'BCKV',
            'ERTA',
            'FLTA',
            'PSTE',
            'VALD',
            'YTDA',
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'OPCL': (
            'ACCC',
            'ACCO',
            'ACCT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OTHR': ('OTHR', 'NTAV'),
    },
    'CAMT': {
        'ACCB': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'DSBR',
            'CAJT',
            'XBRD',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'ODFT',
            'RIMB',
            'SWEP',
            'TAXE',
            'TOPG',
            'ZABA',
        ),
        'CAPL': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'XBRD',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'OTHR': ('OTHR', 'NTAV'),
    },
    'CMDT': {
        'DLVR': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'FTUR': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'OPTN': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OTHR': ('OTHR', 'NTAV'),
        'SPOT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
    },
    'DERV': {
        'LFUT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'LOPT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'OBND': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OCRD': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OEQT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OIRT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OSED': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OSWP': (
            'CHRG',
            'SWCC',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'SWFP',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'SWPP',
            'RIMB',
            'SWRS',
            'TAXE',
            'SWUF',
        ),
        'OTHR': ('OTHR', 'NTAV'),
    },
    'XTND': {'NTAV': ('NTAV',)},
    'FORX': {
        'FWRD': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'FTUR': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NDFX': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'OTHR': ('OTHR', 'NTAV'),
        'SPOT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'SWAP': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
    },
    'LDAS': {
        'CSLN': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DDWN',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'PPAY',
            'RIMB',
            'RNEW',
            'TAXE',
        ),
        'FTDP': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DPST',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'RPMT',
            'TAXE',
        ),
        'FTLN': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DDWN',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'PPAY',
            'RIMB',
            'RNEW',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MGLN': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DDWN',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'PPAY',
            'RIMB',
            'RNEW',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'NTDP': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DPST',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'RPMT',
            'TAXE',
        ),
        'NTLN': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DDWN',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'PPAY',
            'RIMB',
            'RNEW',
            'TAXE',
        ),
        'OTHR': ('OTHR', 'NTAV'),
        'SYDN': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DDWN',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'PPAY',
            'RIMB',
            'RNEW',
            'TAXE',
        ),
    },
    'PMNT': {
        'CNTR': (
            'BCDP',
            'BCWD',
            'CDPT',
            'CWDL',
            'CHRG',
            'CHKD',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'FCDP',
            'FCWD',
            'INTR',
            'MSCD',
            'MIXD',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
            'TCDP',
            'TCWD',
        ),
        'CCRD': (
            'CDPT',
            'CWDL',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'POSC',
            'XBCW',
            'XBCP',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'POSD',
            'RIMB',
            'SMRT',
            'TAXE',
        ),
        'DRFT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'DDFT',
            'UDFT',
            'DMCG',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'STAM',
            'STLR',
            'TAXE',
        ),
        'ICCN': (
            'ACON',
            'BACT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'COAT',
            'CAJT',
            'XICT',
            'DAJT',
            'FEES',
            'FIOA',
            'INTR',
            'ICCT',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'ICHQ': (
            'ARPD',
            'BCHQ',
            'CASH',
            'CSHA',
            'CCCH',
            'CHRG',
            'CCHQ',
            'CQRV',
            'URCQ',
            'CLCQ',
            'COMM',
            'COME',
            'COMI',
            'CDIS',
            'CAJT',
            'CRCQ',
            'DAJT',
            'FEES',
            'XBCQ',
            'XRCQ',
            'INTR',
            'NPCC',
            'COMT',
            'NTAV',
            'OPCQ',
            'ORCQ',
            'OTHR',
            'RIMB',
            'TAXE',
            'UPCQ',
            'XPCQ',
        ),
        'ICDT': (
            'ACOR',
            'ACDT',
            'ADBT',
            'APAC',
            'ARET',
            'AREV',
            'ASET',
            'ATXN',
            'AUTT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'VCOM',
            'XBCT',
            'XBSA',
            'XBST',
            'DAJT',
            'DMCT',
            'FEES',
            'FICT',
            'INTR',
            'BOOK',
            'COMT',
            'ENCT',
            'NTAV',
            'OTHR',
            'SALA',
            'PRCT',
            'RIMB',
            'RPCR',
            'RRTN',
            'XRTN',
            'SDVA',
            'ESCT',
            'STDO',
            'TAXE',
            'TTLS',
        ),
        'IDDT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'XBDD',
            'DAJT',
            'PMDD',
            'URDD',
            'FEES',
            'FIDD',
            'INTR',
            'COMT',
            'NTAV',
            'OODD',
            'OTHR',
            'PADD',
            'RIMB',
            'RCDD',
            'PRDD',
            'UPDD',
            'BBDD',
            'ESDD',
            'TAXE',
        ),
        'IRCT': (
            'ACOR',
            'ACDT',
            'ADBT',
            'APAC',
            'ARET',
            'AREV',
            'ASET',
            'ATXN',
            'AUTT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'VCOM',
            'XBCT',
            'XBSA',
            'XBST',
            'DAJT',
            'DMCT',
            'FEES',
            'FICT',
            'INTR',
            'BOOK',
            'COMT',
            'ENCT',
            'NTAV',
            'OTHR',
            'SALA',
            'PRCT',
            'RIMB',
            'RPCR',
            'RRTN',
            'XRTN',
            'SDVA',
            'ESCT',
            'STDO',
            'TAXE',
            'TTLS',
        ),
        'LBOX': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'LBCA',
            'CAJT',
            'LBDB',
            'DAJT',
            'LBDP',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCRD': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'POSC',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'POSP',
            'RIMB',
            'SMCD',
            'TAXE',
            'UPCT',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
            'IADD',
        ),
        'NTAV': ('NTAV',),
        'OTHR': ('OTHR', 'NTAV'),
        'RCCN': (
            'ACON',
            'BACT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'COAT',
            'CAJT',
            'XICT',
            'DAJT',
            'FEES',
            'FIOA',
            'INTR',
            'ICCT',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'RCHQ': (
            'ARPD',
            'BCHQ',
            'CASH',
            'CSHA',
            'CCCH',
            'CHRG',
            'CCHQ',
            'CQRV',
            'URCQ',
            'CLCQ',
            'COMM',
            'COME',
            'COMI',
            'CDIS',
            'CAJT',
            'CRCQ',
            'DAJT',
            'FEES',
            'XBCQ',
            'XRCQ',
            'INTR',
            'NPCC',
            'COMT',
            'NTAV',
            'OPCQ',
            'ORCQ',
            'OTHR',
            'RIMB',
            'TAXE',
            'UPCQ',
            'XPCQ',
        ),
        'RCDT': (
            'ACOR',
            'ACDT',
            'ADBT',
            'APAC',
            'ARET',
            'AREV',
            'ASET',
            'ATXN',
            'AUTT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'VCOM',
            'XBCT',
            'XBSA',
            'XBST',
            'DAJT',
            'DMCT',
            'FEES',
            'FICT',
            'INTR',
            'BOOK',
            'COMT',
            'ENCT',
            'NTAV',
            'OTHR',
            'SALA',
            'PRCT',
            'RIMB',
            'RPCR',
            'RRTN',
            'XRTN',
            'SDVA',
            'ESCT',
            'STDO',
            'TAXE',
            'TTLS',
        ),
        'RDDT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'XBDD',
            'DAJT',
            'PMDD',
            'URDD',
            'FEES',
            'FIDD',
            'INTR',
            'COMT',
            'NTAV',
            'OODD',
            'OTHR',
            'PADD',
            'RIMB',
            'RCDD',
            'PRDD',
            'UPDD',
            'BBDD',
            'ESDD',
            'TAXE',
        ),
        'RRCT': (
            'ACOR',
            'ACDT',
            'ADBT',
            'APAC',
            'ARET',
            'AREV',
            'ASET',
            'ATXN',
            'AUTT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'VCOM',
            'XBCT',
            'XBSA',
            'XBST',
            'DAJT',
            'DMCT',
            'FEES',
            'FICT',
            'INTR',
            'BOOK',
            'COMT',
            'ENCT',
            'NTAV',
            'OTHR',
            'SALA',
            'PRCT',
            'RIMB',
            'RPCR',
            'RRTN',
            'XRTN',
            'SDVA',
            'ESCT',
            'STDO',
            'TAXE',
            'TTLS',
        ),
    },
    'PMET': {
        'DLVR': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'FTUR': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'OPTN': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'OTHR': ('OTHR', 'NTAV'),
        'SPOT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
    },
    'SECU': {
        'BLOC': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'XCHG',
            'XCHC',
            'XCHN',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTCG',
            'OTCC',
            'OTCN',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'COLL': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CMBO',
            'CMCO',
            'CPRB',
            'CAJT',
            'DAJT',
            'EQBO',
            'EQCO',
            'FEES',
            'FWBC',
            'FWCC',
            'INTR',
            'SLBC',
            'SLCC',
            'MGCC',
            'MARG',
            'COMT',
            'NTAV',
            'OPBC',
            'OPCC',
            'OTHR',
            'RIMB',
            'REPU',
            'SECB',
            'SECL',
            'SWBC',
            'TAXE',
            'TRPO',
        ),
        'CORP': (
            'BONU',
            'EXRI',
            'CAPG',
            'DVCA',
            'CSLI',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CONV',
            'CAJT',
            'DAJT',
            'DECR',
            'DVOP',
            'DRIP',
            'DRAW',
            'DTCH',
            'SHPR',
            'EXOF',
            'FEES',
            'REDM',
            'MCAL',
            'INTR',
            'PRII',
            'INTR',
            'LIQU',
            'MRGR',
            'COMT',
            'NTAV',
            'ODLT',
            'OTHR',
            'PCAL',
            'PRED',
            'PRIO',
            'BPUT',
            'RWPL',
            'RIMB',
            'BIDS',
            'RHTS',
            'SSPL',
            'TREC',
            'TAXE',
            'TEND ',
            'EXWA',
        ),
        'OTHB': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'XCHG',
            'XCHC',
            'XCHN',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTCG',
            'OTCC',
            'OTCN',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'CUST': (
            'BONU',
            'EXRI',
            'CAPG',
            'DVCA',
            'CSLI',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CONV',
            'CAJT',
            'DAJT',
            'DECR',
            'DVOP',
            'DRIP',
            'DRAW',
            'DTCH',
            'SHPR',
            'EXOF',
            'FEES',
            'REDM',
            'MCAL',
            'INTR',
            'PRII',
            'INTR',
            'LIQU',
            'MRGR',
            'COMT',
            'NTAV',
            'ODLT',
            'OTHR',
            'PCAL',
            'PRED',
            'PRIO',
            'BPUT',
            'RWPL',
            'RIMB',
            'BIDS',
            'RHTS',
            'SSPL',
            'TREC',
            'TAXE',
            'TEND ',
            'EXWA',
        ),
        'COLC': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'LACK': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'CASH': (
            'BKFE',
            'ERWI',
            'BROK',
            'CPEN',
            'CHAR',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CLAI',
            'CAJT',
            'DAJT',
            'GEN2',
            'FEES',
            'INFD',
            'FUTU',
            'FUCO',
            'RESI',
            'PRIN',
            'INTR',
            'ERWA',
            'MNFE',
            'COMT',
            'NTAV',
            'OTHR',
            'OVCH',
            'RIMB',
            'STAM',
            'SWAP',
            'SWEP',
            'TREC',
            'TAXE',
            'TRFE',
            'UNCO',
            'GEN1',
            'WITH',
        ),
        'NSET': (
            'BSBO',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'CROS',
            'DAJT',
            'ISSU',
            'XCHG',
            'XCHC',
            'XCHN',
            'OWNE',
            'FCTA',
            'FEES',
            'INSP',
            'INTR',
            'OWNI',
            'NETT',
            'NSYN',
            'COMT',
            'NTAV',
            'OTCG',
            'OTCC',
            'OTCN',
            'OTHR',
            'PAIR',
            'PLAC',
            'PORT',
            'PRUD',
            'REDM',
            'REAA',
            'RIMB',
            'REPU',
            'RVPO',
            'SECB',
            'SECL',
            'BSBC',
            'SUBS',
            'SUAA',
            'SWIC',
            'SYND',
            'TAXE',
            'TBAC',
            'TRAD',
            'TRIN',
            'TOUT',
            'TRPO',
            'TRVO',
            'TURN',
        ),
        'NTAV': ('NTAV',),
        'OTHR': ('OTHR', 'NTAV'),
        'SETT': (
            'BSBO',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'CROS',
            'DAJT',
            'ISSU',
            'XCHG',
            'XCHC',
            'XCHN',
            'OWNE',
            'FCTA',
            'FEES',
            'INSP',
            'INTR',
            'OWNI',
            'NETT',
            'NSYN',
            'COMT',
            'NTAV',
            'OTCG',
            'OTCC',
            'OTCN',
            'OTHR',
            'PAIR',
            'PLAC',
            'PORT',
            'PRUD',
            'REDM',
            'REAA',
            'RIMB',
            'REPU',
            'RVPO',
            'SECB',
            'SECL',
            'BSBC',
            'SUBS',
            'SUAA',
            'SWIC',
            'SYND',
            'TAXE',
            'TBAC',
            'TRAD',
            'TRIN',
            'TOUT',
            'TRPO',
            'TRVO',
            'TURN',
        ),
    },
    'TRAD': {
        'CLNC': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'STAC',
            'STLR',
            'TAXE',
        ),
        'DOCC': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'STAC',
            'STLR',
            'TAXE',
        ),
        'DCCT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'FRZF',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'SABG',
            'SOSE',
            'SOSI',
            'STLR',
            'TAXE',
        ),
        'GUAR': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'STLM',
            'TAXE',
        ),
        'MCOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'MDOP': (
            'ADJT',
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'FEES',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'TAXE',
        ),
        'NTAV': ('NTAV',),
        'OTHR': ('OTHR', 'NTAV'),
        'LOCT': (
            'CHRG',
            'COMM',
            'COME',
            'COMI',
            'CAJT',
            'DAJT',
            'FEES',
            'FRZF',
            'INTR',
            'COMT',
            'NTAV',
            'OTHR',
            'RIMB',
            'SABG',
            'SOSE',
            'SOSI',
            'STLR',
            'TAXE',
        ),
    },
}

# Some stet banks don't follow ISO20022 standard but follow a CFONB format
# See: https://www.cfonb.org/fichiers/20190913092943_Brochure_Codes_Operation_pour_restitutions_clienteles_V5_0_novembre_2018.pdf
FRENCH_BANK_TRANSACTION_CODES = {
    '01': ('PMNT', 'ICHQ', 'CCHQ'),
    '02': ('PMNT', 'RCHQ', 'CCHQ'),
    '03': ('PMNT', 'RCHQ', 'UPCQ'),
    '04': ('PMNT', 'CCRD', 'CDPT'),
    '05': ('PMNT', 'RCDT', 'DMCT'),
    '06': ('PMNT', 'ICDT', 'DMCT'),
    '07': ('PMNT', 'DRFT', 'STAM'),
    '08': ('PMNT', 'RDDT', 'OTHR'),
    '09': ('PMNT', 'IDDT', 'OTHR'),
    '10': ('PMNT', 'RDDT', 'UPDD'),
    '11': ('PMNT', 'CCRD', 'RIMB'),
    '12': ('PMNT', 'ICDT', 'RRTN'),
    '13': ('PMNT', 'RCCN', 'ICCT'),
    '14': ('PMNT', 'ICCN', 'ICCT'),
    '15': ('PMNT', 'RCHQ', 'CCHQ'),
    # Missing numbers are also missing in their documentation
    '18': ('PMNT', 'RCDT', 'DMCT'),
    '21': ('PMNT', 'ICDT', 'DMCT'),
    '22': ('PMNT', 'RDDT', 'OODD'),
    '24': ('PMNT', 'IDDT', 'OODD'),
    '26': ('PMNT', 'IDDT', 'UPDD'),
    '27': ('PMNT', 'RDDT', 'UPDD'),
    '28': ('PMNT', 'CCRD', 'POSD'),
    '29': ('PMNT', 'CCRD', 'CWDL'),
    '30': ('PMNT', 'MCRD', 'POSC'),
    '31': ('PMNT', 'DRFT', 'STAM'),
    '32': ('PMNT', 'DRFT', 'DDFT'),
    '33': ('PMNT', 'DRFT', 'UDFT'),
    '34': ('PMNT', 'DRFT', 'OTHR'),
    '35': ('PMNT', 'DRFT', 'STAM'),
    '36': ('PMNT', 'MCRD', 'DAJT'),
    '37': ('PMNT', 'DRFT', 'DDFT'),
    '38': ('PMNT', 'MCRD', 'UPCT'),
    '39': ('PMNT', 'RCDT', 'XBCT'),
    '41': ('PMNT', 'ICDT', 'XBCT'),
    '42': ('FORX', 'MDOP', 'OTHR'),
    '43': ('PMNT', 'MDOP', 'OTHR'),
    '44': ('PMNT', 'ICDT', 'XBCT'),
    '45': ('PMNT', 'RCDT', 'XBCT'),
    '46': ('FORX', 'SPOT', 'OTHR'),
    '47': ('FORX', 'SPOT', 'OTHR'),
    '48': ('FORX', 'FWRD', 'OTHR'),
    '49': ('FORX', 'FWRD', 'OTHR'),
    '51': ('SECU', 'SETT', 'SUBS'),
    '52': ('SECU', 'MDOP', 'OTHR'),
    '53': ('SECU', 'SETT', 'SUBS'),
    '54': ('SECU', 'SETT', 'REDM'),
    '55': ('SECU', 'CASH', 'OTHR'),
    '56': ('SECU', 'SETT', 'SUBS'),
    '57': ('SECU', 'SETT', 'REDM'),
    '58': ('SECU', 'SETT', 'SUBS'),
    '59': ('SECU', 'SETT', 'REDM'),
    '61': ('ACMT', 'MDOP', 'INTR'),
    '62': ('ACMT', 'MDOP', 'COMM'),
    '63': ('ACMT', 'MCOP', 'INTR'),
    '64': ('ACMT', 'MDOP', 'COMI'),
    '65': ('ACMT', 'MDOP', 'COME'),
    '66': ('ACMT', 'MDOP', 'COMT'),
    '67': ('ACMT', 'MDOP', 'TAXE'),
    '68': ('FORX', 'MDOP', 'OTHR'),
    '69': ('FORX', 'MCOP', 'OTHR'),
    '70': ('ACMT', 'MDOP', 'INTR'),
    '71': ('LDAS', 'CSLN', 'DDWN'),
    '72': ('LDAS', 'CSLN', 'PPAY'),
    '73': ('SECU', 'CORP', 'OTHR'),
    '74': ('SECU', 'CORP', 'OTHR'),
    '75': ('LDAS', 'CSLN', 'RIMB'),
    '76': ('LDAS', 'FTDP', 'DPST'),
    '77': ('LDAS', 'FTDP', 'RPMT'),
    '80': ('SECU', 'SETT', 'SUBS'),
    '81': ('SECU', 'SETT', 'REDM'),
    '82': ('SECU', 'SETT', 'REDM'),
    '83': ('SECU', 'SETT', 'REDM'),
    '84': ('SECU', 'SETT', 'SUBS'),
    '85': ('SECU', 'SETT', 'REDM'),
    '86': ('SECU', 'MDOP', 'OTHR'),
    '87': ('SECU', 'CUST', 'DVCA'),
    '88': ('SECU', 'CUST', 'COMM'),
    '89': ('SECU', 'CUST', 'TREC'),
    '90': ('PMNT', 'ICDT', 'IADD'),
    '91': ('ACMT', 'MCOP', 'OTHR'),
    '92': ('CAMT', 'ACCB', 'OTHR'),
    '93': ('CAMT', 'ACCB', 'OTHR'),
    '94': ('CAMT', 'CAPL', 'OTHR'),
    '95': ('CAMT', 'CAPL', 'OTHR'),
    '96': ('PMNT', 'RDDT', 'PADD'),
    '97': ('PMNT', 'IDDT', 'PADD'),
    '98': ('PMNT', 'IDDT', 'UPDD'),
    '99': ('ACMT', 'MCOP', 'ADJT'),
    'A1': ('PMNT', 'IDDT', 'ESDD'),
    'A2': ('PMNT', 'IDDT', 'BBDD'),
    'A3': ('PMNT', 'RDDT', 'UPDD'),
    'A4': ('PMNT', 'RDDT', 'UPDD'),
    'A5': ('PMNT', 'IDDT', 'PRDD'),
    'A6': ('PMNT', 'IDDT', 'PRDD'),
    'B1': ('PMNT', 'RDDT', 'ESDD'),
    'B2': ('PMNT', 'RDDT', 'BBDD'),
    'B3': ('PMNT', 'IDDT', 'UPDD'),
    'B4': ('PMNT', 'IDDT', 'UPDD'),
    'B5': ('PMNT', 'RDDT', 'PRDD'),
    'B6': ('PMNT', 'RDDT', 'PRDD'),
    'C1': ('PMNT', 'IRCT', 'ESCT'),
    'C2': ('PMNT', 'RRCT', 'ESCT'),
    'C3': ('PMNT', 'IRCT', 'RPCR'),
    'C4': ('PMNT', 'RRCT', 'RPCR'),
    'C5': ('PMNT', 'IRCT', 'RIMB'),
}



# Copyright(C) 2013 Pierre Mazière
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.capabilities.base import BaseObject, StringField, NotLoaded

__all__ = ['StreamInfo']


class StreamInfo(BaseObject):
    """
    Stream related information.
    """
    who = StringField('Who is currently on air')
    what = StringField('What is currently on air')

    def __iscomplete__(self):
        return self.who is not NotLoaded or self.what is not NotLoaded

    def __str__(self):
        if self.who:
            return '%s - %s' % (self.who, self.what)
        else:
            return self.what



# Copyright(C) 2010-2011 Romain Bignon, Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from threading import Event, RLock
from threading import Timer

from woob.tools.log import getLogger
from woob.tools.misc import get_backtrace


__all__ = ['Scheduler']


class IScheduler:
    """Interface of a scheduler."""

    def schedule(self, interval, function, *args):
        """
        Schedule an event.

        :param interval: delay before calling the function
        :type interval: int
        :param function: function to call
        :type function: callabale
        :param args: arguments to give to function
        :returns: an event identificator
        """
        raise NotImplementedError()

    def repeat(self, interval, function, *args):
        """
        Repeat a call to a function

        :param interval: interval between two calls
        :type interval: int
        :param function: function to call
        :type function: callable
        :param args: arguments to give to function
        :returns: an event identificator
        """
        raise NotImplementedError()

    def cancel(self, ev):
        """
        Cancel an event

        :param ev: the event identificator
        """

        raise NotImplementedError()

    def run(self):
        """
        Run the scheduler loop
        """
        raise NotImplementedError()

    def want_stop(self):
        """
        Plan to stop the scheduler.
        """
        raise NotImplementedError()


class RepeatedTimer(Timer):
    def run(self):
        while not self.finished.is_set():
            try:
                self.function(*self.args, **self.kwargs)
            except Exception:
                # do not stop timer because of an exception
                print(get_backtrace())
            self.finished.wait(self.interval)
        self.finished.set()


class Scheduler(IScheduler):
    """Scheduler using Python's :mod:`threading`."""

    def __init__(self):
        self.logger = getLogger('%s.scheduler' % __name__)
        self.mutex = RLock()
        self.stop_event = Event()
        self.count = 0
        self.queue = {}

    def schedule(self, interval, function, *args):
        return self._schedule(Timer, interval, self._schedule_callback, function, *args)

    def repeat(self, interval, function, *args):
        return self._schedule(RepeatedTimer, interval, self._repeat_callback, function, *args)

    def _schedule(self, klass, interval, meta_func, function, *args):
        if self.stop_event.is_set():
            return

        with self.mutex:
            self.count += 1
            self.logger.debug('function "%s" will be called in %s seconds' % (function.__name__, interval))
            timer = klass(interval, meta_func, (self.count, interval, function, args))
            self.queue[self.count] = timer
            timer.start()
            return self.count

    def _schedule_callback(self, count, interval, function, args):
        with self.mutex:
            self.queue.pop(count)
        return function(*args)

    def _repeat_callback(self, count, interval, function, args):
        function(*args)
        with self.mutex:
            try:
                e = self.queue[count]
            except KeyError:
                return
            else:
                self.logger.debug('function "%s" will be called in %s seconds' % (function.__name__, e.interval))

    def cancel(self, ev):
        with self.mutex:
            try:
                e = self.queue.pop(ev)
            except KeyError:
                return False
            e.cancel()
            self.logger.debug('scheduled function "%s" is canceled' % e.function.__name__)
            return True

    def _wait_to_stop(self):
        self.want_stop()
        with self.mutex:
            for e in self.queue.values():
                e.cancel()
                e.join()
            self.queue = {}

    def run(self):
        try:
            while True:
                self.stop_event.wait(0.1)
        except KeyboardInterrupt:
            self._wait_to_stop()
            raise
        else:
            self._wait_to_stop()
        return True

    def want_stop(self):
        self.stop_event.set()
        with self.mutex:
            for t in self.queue.values():
                t.cancel()
                # Contrary to _wait_to_stop(), don't call t.join
                # because want_stop() have to be non-blocking.
            self.queue = {}



# Copyright(C) 2013  Florent Fourcot
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.base import empty
from woob.capabilities.gauge import CapGauge, SensorNotFound
from woob.tools.application.repl import ReplApplication
from woob.tools.application.formatters.iformatter import IFormatter


__all__ = ['AppGauge']


class GaugeFormatter(IFormatter):
    MANDATORY_FIELDS = ('name', 'object', 'sensors')
    DISPLAYED_FIELDS = ('city', )

    def start_format(self, **kwargs):
        # Name = 27   Object = 10   City = 10  Sensors = 33
        self.output(' Name and ID                  Object     City       Sensors                         ')
        self.output('----------------------------+----------+----------+---------------------------------')

    def format_obj(self, obj, alias):
        name = obj.name
        city = ""
        if not empty(obj.city):
            city = obj.city

        if not obj.sensors or (len(obj.sensors) == 0):
            result = ' %s %s %s \n' %\
                   (self.colored('%-27s' % name[:27], 'red'),
                    self.colored('%-10s' % obj.object[:10], 'yellow'),
                    self.colored('%-10s' % city[:10], 'yellow')
                    )
            result += ' %s \n' % self.colored('%-47s' % obj.fullid[:47], 'blue')
        else:
            first = True
            firstaddress = obj.sensors[0].address
            for sensor in obj.sensors:
                sensorname = sensor.name
                # This is a int value, do not display it as a float
                if not empty(sensor.lastvalue.level):
                    if int(sensor.lastvalue.level) == sensor.lastvalue.level:
                        lastvalue = "%d " % sensor.lastvalue.level
                    else:
                        lastvalue = "%r " % sensor.lastvalue.level
                    if not empty(sensor.unit):
                        lastvalue += "%s" % sensor.unit
                else:
                    lastvalue = "? "
                if first:
                    result = ' %s %s %s ' %\
                             (self.colored('%-27s' % name[:27], 'red'),
                              self.colored('%-10s' % obj.object[:10], 'yellow'),
                              self.colored('%-10s' % city[:10], 'yellow'),
                              )
                    if not empty(firstaddress):
                        result += '%s' % self.colored('%-33s' % sensor.address[:33], 'yellow')
                    result += '\n'
                    result += ' %s' % self.colored('%-47s' % obj.fullid[:47], 'blue')
                    result += '   %s %s\n' %\
                              (self.colored('%-20s' % sensorname[:20], 'magenta'),
                               self.colored('%-13s' % lastvalue[:13], 'red')
                               )
                    first = False
                else:
                    result += '                                                   %s %s\n' %\
                              (self.colored('%-20s' % sensorname[:20], 'magenta'),
                               self.colored('%-13s' % lastvalue[:13], 'red')
                               )
                    if not empty(sensor.address) and sensor.address != firstaddress:
                        result += '                                                   %s \n' %\
                                  self.colored('%-33s' % sensor.address[:33], 'yellow')

        return result


class AppGauge(ReplApplication):
    APPNAME = 'gauge'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2013-YEAR Florent Fourcot'
    DESCRIPTION = "Console application allowing to display various sensors and gauges values."
    SHORT_DESCRIPTION = "display sensors and gauges values"
    CAPS = (CapGauge)
    DEFAULT_FORMATTER = 'table'
    EXTRA_FORMATTERS = {'gauge_list':   GaugeFormatter, }
    COMMANDS_FORMATTERS = {'search':    'gauge_list',
                           }

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    def bcall_error_handler(self, backend, error, backtrace):
        if isinstance(error, SensorNotFound):
            msg = str(error) or 'Sensor not found (hint: try details command)'
            print('Error(%s): %s' % (backend.name, msg), file=self.stderr)
        else:
            return super().bcall_error_handler(backend, error, backtrace)

    def do_search(self, pattern):
        """
        search [PATTERN]

        Display all gauges. If PATTERN is specified, search on a pattern.
        """
        self.change_path(['gauges'])
        self.start_format()
        for gauge in self.do('iter_gauges', pattern or None, caps=CapGauge):
            self.cached_format(gauge)

    def complete_search(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_details(self, line):
        """
        details GAUGE_ID

        Display details of all sensors of the gauge.
        """
        gauge, pattern = self.parse_command_args(line, 2, 1)
        _id, backend_name = self.parse_id(gauge)

        self.start_format()
        for sensor in self.do('iter_sensors', _id, pattern=pattern, backends=backend_name, caps=CapGauge):
            self.format(sensor)

    def do_history(self, line):
        """
        history SENSOR_ID

        Get history of a specific sensor (use 'search' to find a gauge, and sensors GAUGE_ID to list sensors attached to the gauge).
        """
        gauge, = self.parse_command_args(line, 1, 1)
        _id, backend_name = self.parse_id(gauge)

        self.start_format()
        for measure in self.do('iter_gauge_history', _id, backends=backend_name, caps=CapGauge):
            self.format(measure)

    def complete_last_sensor_measure(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_last_sensor_measure(self, line):
        """
        last_sensor_measure SENSOR_ID

        Get last measure of a sensor.
        """
        gauge, = self.parse_command_args(line, 1, 1)
        _id, backend_name = self.parse_id(gauge)

        self.start_format()
        for measure in self.do('get_last_measure', _id, backends=backend_name, caps=CapGauge):
            self.format(measure)



# Copyright(C) 2010-2015 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import datetime
import time

from .base import Capability, BaseObject, NotLoaded, Field, StringField, \
                  IntField, UserError
from .date import DateField

__all__ = ['Thread', 'Message', 'CapMessages', 'CantSendMessage', 'CapMessagesPost']


class Message(BaseObject):
    """
    Represents a message read or to send.
    """
    IS_HTML = 0x001
    "The content is HTML formatted"
    IS_UNREAD = 0x002
    "The message is unread"
    IS_RECEIVED = 0x004
    "The receiver has read this message"
    IS_NOT_RECEIVED = 0x008
    "The receiver hass not read this message"

    thread =        Field('Reference to the thread', 'Thread')
    title =         StringField('Title of message')
    sender =        StringField('Author of this message')
    receivers =     Field('Receivers of the message', list)
    date =          DateField('Date when the message has been sent')
    content =       StringField('Body of message')
    signature =     StringField('Optional signature')
    parent =        Field('Parent message', 'Message')
    children =      Field('Children fields', list)
    flags =         IntField('Flags (IS_* constants)', default=0)

    def __init__(self, thread=NotLoaded,
                       id=NotLoaded,
                       title=NotLoaded,
                       sender=NotLoaded,
                       receivers=NotLoaded,
                       date=None,
                       parent=NotLoaded,
                       content=NotLoaded,
                       signature=NotLoaded,
                       children=NotLoaded,
                       flags=0,
                       url=None):
        super(Message, self).__init__(id, url)
        self.thread = thread
        self.title = title
        self.sender = sender
        self.receivers = receivers
        self.content = content
        self.signature = signature
        self.children = children
        self.flags = flags

        if date is None:
            date = datetime.datetime.utcnow()
        self.date = date

        if isinstance(parent, Message):
            self.parent = parent
        else:
            self.parent = NotLoaded
            self._parent_id = parent

    @property
    def date_int(self):
        """
        Date of message as an integer.
        """
        return int(time.strftime('%Y%m%d%H%M%S', self.date.timetuple()))

    @property
    def full_id(self):
        """
        Full ID of message (in form '**THREAD_ID.MESSAGE_ID**')
        """
        return '%s.%s' % (self.thread.id, self.id)

    @property
    def full_parent_id(self):
        """
        Get the full ID of the parent message (in form '**THREAD_ID.MESSAGE_ID**').
        """
        if self.parent:
            return self.parent.full_id
        elif self._parent_id is None:
            return ''
        elif self._parent_id is NotLoaded:
            return NotLoaded
        else:
            return '%s.%s' % (self.thread.id, self._parent_id)

    def __eq__(self, msg):
        if not isinstance(msg, Message):
            return False

        if self.thread:
            return str(self.thread.id) == str(msg.thread.id) and \
                   str(self.id) == str(msg.id)
        else:
            return str(self.id) == str(msg.id)

    def __repr__(self):
        return '<Message id=%r title=%r date=%r from=%r>' % (
                   self.full_id, self.title, self.date, self.sender)


class Thread(BaseObject):
    """
    Thread containing messages.
    """
    IS_THREADS =    0x001
    IS_DISCUSSION = 0x002

    root =      Field('Root message', Message)
    title =     StringField('Title of thread')
    date =      DateField('Date of thread')
    flags =     IntField('Flags (IS_* constants)', default=IS_THREADS)

    def iter_all_messages(self):
        """
        Iter all messages of the thread.

        :rtype: iter[:class:`Message`]
        """
        if self.root:
            yield self.root
            for m in self._iter_all_messages(self.root):
                yield m

    def _iter_all_messages(self, message):
        if message.children:
            for child in message.children:
                yield child
                for m in self._iter_all_messages(child):
                    yield m


class CapMessages(Capability):
    """
    Capability to read messages.
    """

    def iter_threads(self):
        """
        Iterates on threads, from newers to olders.

        :rtype: iter[:class:`Thread`]
        """
        raise NotImplementedError()

    def get_thread(self, id):
        """
        Get a specific thread.

        :rtype: :class:`Thread`
        """
        raise NotImplementedError()

    def iter_unread_messages(self):
        """
        Iterates on messages which hasn't been marked as read.

        :rtype: iter[:class:`Message`]
        """
        raise NotImplementedError()

    def set_message_read(self, message):
        """
        Set a message as read.

        :param message: message read (or ID)
        :type message: :class:`Message` or str
        """
        raise NotImplementedError()


class CantSendMessage(UserError):
    """
    Raised when a message can't be send.
    """


class CapMessagesPost(Capability):
    """
    This capability allow user to send a message.
    """

    def post_message(self, message):
        """
        Post a message.

        :param message: message to send
        :type message: :class:`Message`
        :raises: :class:`CantSendMessage`
        """
        raise NotImplementedError()
